"""
Priors
======

Learn how you can include prior knowledge into the fitting by setting
priors on single parameters.

Prerequisites
-------------

-  Knowledge of spectral analysis to produce 1D spectral datasets, see
   the :doc:`/tutorials/analysis-1d/spectral_analysis` tutorial.
-  Knowledge of fitting models to datasets, see
   :doc:`/tutorials/details/fitting` tutorial
-  General knowledge of statistics and priors

Context
-------

Generally, the set of parameters describing best the data is where the
fit statistic :math:`2 x log L` has its global minimum. Depending on
the type of background estimation, it is either the Cash fit statistic
or the WStat (see :doc:`/user-guide/stats/fit_statistics` for more
details).

A prior on the value of some of the parameters can be added to this fit
statistic. The prior is again a probability density function of the
model parameters and can take different forms, including Gaussian
distributions, uniform distributions, etc. The prior includes
information or knowledge about the dataset or the parameters of the fit.

Setting priors on multiple parameters simultaneously is supported,
However, for now, they should not be correlated.

The spectral dataset used here contains a simulated power-law source and
its IRFs are based on H.E.S.S. data of the Crab Nebula (similar to
:doc:`/tutorials/details/fitting`). We are simulating the source here, so
that we can control the input and check the correctness of the fit
results.

The tutorial addresses three examples:

-  Including prior information about the sources index
-  How to add a custom prior class?
-  Encouraging positive amplitude values

In the first example, the Gaussian prior is used. It is shown how to set
a prior on a model parameter and how it modifies the fit statistics. A
source is simulated without statistics and fitted with and without the
priors. The different fit results are discussed.

For the second example, 1000 datasets containing a very weak source are
simulated. Due to the statistical fluctuations, the amplitude’s best-fit
value is negative for some draws. By setting a uniform prior on the
amplitude, this can be avoided.

The setup
---------

"""

import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt
from gammapy.datasets import SpectrumDatasetOnOff
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    GaussianPrior,
    PowerLawSpectralModel,
    SkyModel,
)


######################################################################
# Model and dataset
# -----------------
#
# First, we define the source model, a power-law with an index of
# :math:`2.3`
#

pl_spectrum = PowerLawSpectralModel(
    index=2.3,
    amplitude=1e-11 / u.cm**2 / u.s / u.TeV,
)
model = SkyModel(spectral_model=pl_spectrum, name="simu-source")


######################################################################
# The data and background are read from pre-computed ON/OFF datasets of
# H.E.S.S. observations. For simplicity, we stack them together and transform
# the dataset to a `~gammapy.datasets.SpectrumDataset`. Then we set the model and create
# an Asimov dataset (dataset without statistics) by setting the counts as
# the model predictions.
#

dataset = SpectrumDatasetOnOff.read(
    "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits"
)

# Set model and fit range
e_min = 0.66 * u.TeV
e_max = 30 * u.TeV
dataset.mask_fit = dataset.counts.geom.energy_mask(e_min, e_max)
dataset = dataset.to_spectrum_dataset()

dataset1 = dataset.copy()
dataset1.models = model.copy(name="model")
dataset1.counts = dataset1.npred()


######################################################################
# Example 1: Including Prior Information about the Sources Index
# --------------------------------------------------------------
#
# This example illustrate the maximum a posteriori (MAP) method which
# differs from the maximum likelihood by the introduction of priors.
# The index was assumed to be :math:`2.3`. However, let us assume you
# have reasons to believe that the index value of the source is actually
# :math:`2.1`. This can be due to theoretical predictions, other
# instruments’ results, etc. We can now create a Gaussian distributed
# prior with a minimum at the expected value of :math:`2.1`. The standard
# deviation of the Gaussian quantifies how much we believe the prior
# knowledge to be true. The smaller the standard deviation, the stronger
# the constraining ability of the prior. For now, we set it to the
# arbitrary value of :math:`0.1`.
#

# initialising the gaussian prior
gaussianprior = GaussianPrior(mu=2.1, sigma=0.1)
# setting the gaussian prior on the index parameter
model_prior = model.copy()
model_prior.parameters["index"].prior = gaussianprior


######################################################################
# The value of the prior depends on the value of the index. If the index
# value equals the Gaussians mean (here 2.1), the prior is zero. This
# means that nothing is added to the cash statistics, and this value is
# favoured in the fit. If the value of the index deviates from the mean of
# 2.1, a prior value > 0 is added to the cash statistics.
#

# For the visualisation, the values are appended to the list; this is not a necessity for the fitting
prior_stat_sums = []
with model_prior.parameters.restore_status():
    i_scan = np.linspace(1.8, 2.4, 100)
    for a in i_scan:
        model_prior.parameters["index"].value = a
        prior_stat_sums.append(model_prior.parameters.prior_stat_sum())

plt.plot(
    i_scan,
    prior_stat_sums,
    color="tab:orange",
    linestyle="dashed",
    label=f"Gaussian Prior: \n$\mu = {gaussianprior.mu.value}$; $\sigma = {gaussianprior.sigma.value}$",
)
plt.axvline(x=gaussianprior.mu.value, color="red")

plt.xlabel("Index Value")
plt.ylabel("Prior")
plt.legend()
plt.show()


######################################################################
# Fitting a Dataset with and without the Prior
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# Now, we copy the dataset consisting of the power-law source and assign
# the model with the Gaussian prior set on its index to it. Both of the
# datasets are fitted.
#
dataset1.fake()
dataset1_prior = dataset1.copy()
dataset1_prior.models = model_prior.copy(name="prior-model")

fit = Fit()
results = fit.run(dataset1)

results_prior = fit.run(dataset1_prior)


######################################################################
# The parameters table will mention the type of prior associated to each model

print(results_prior.models.to_parameters_table())


######################################################################
# To see the details of the priors, eg:

print(results_prior.models.parameters["index"].prior)

######################################################################
# The Likelihood profiles can be computed for both the datasets. Hereby,
# the likelihood gets computed for different values of the index. For each
# step the other free parameters are getting reoptimized.
#

dataset1.models.parameters["index"].scan_n_values = 20
dataset1_prior.models.parameters["index"].scan_n_values = 20

scan = fit.stat_profile(datasets=dataset1, parameter="index", reoptimize=True)

scan_prior = fit.stat_profile(
    datasets=dataset1_prior, parameter="index", reoptimize=True
)


######################################################################
# Now, we can compare the two Likelihood scans. In the first case, we did
# not set any prior. This means that only the Cash Statistic itself is
# getting minimized. The Cash statistics minimum is the actual value of
# the index we used for the simulation (:math:`2.3`). Therefore, the
# best-fit value was found to be :math:`2.3`. Note how the error bars
# correspond to the :math:`1\sigma` error, i.e. where the stat sum equals
# the minimum + 1.
#
# The plot also shows the prior we set on the index for the second
# dataset. The scan was computed above. If the logarithm of the prior is added to the cash
# statistics, one ends up with the posterior function. The posterior is
# minimized during the second fit to obtain the maximum a posteriori estimation. Its minimum is between the priors and
# the cash statistics minimum. The best-fit value is :math:`2.16`. We
# weighed the truth with our prior beliefs and ended up with a compromise
# between the values. The uncertainty of the parameter is again where the
# posterior distributions equal its minimum + 1.
#
# The best-fit index value and uncertainty depend strongly on the standard
# deviation of the Gaussian prior. You can vary :math:`\sigma` and see how
# the Likelihood profiles and corresponding minima will change.
#

plt.plot(scan["model.spectral.index_scan"], scan["stat_scan"], label="Cash Statistics")
plt.plot(
    i_scan,
    np.array(prior_stat_sums) + np.min(scan["stat_scan"]),
    linestyle="dashed",
    label="Prior",
)
plt.plot(
    scan_prior["prior-model.spectral.index_scan"],
    scan_prior["stat_scan"],
    label="Posterior\n(Cash Stat. + Prior)",
)

par = dataset1.models.parameters["index"]
plt.errorbar(
    x=par.value,
    y=np.min(scan["stat_scan"]) + 1,
    xerr=par.error,
    fmt="x",
    markersize=6,
    capsize=8,
    color="darkblue",
    label="Best-Fit w/o Prior",
)
par = dataset1_prior.models.parameters["index"]
plt.errorbar(
    x=par.value,
    y=np.min(scan_prior["stat_scan"]) + 1,
    xerr=par.error,
    fmt="x",
    markersize=6,
    capsize=8,
    color="darkgreen",
    label="Best-Fit w/ Prior",
)
plt.legend()
# plt.ylim(31.5,35)
plt.xlabel("Index")
plt.ylabel("Fit Statistics [arb. unit]")
plt.show()


######################################################################
# This example shows a critical note on using priors: we were able to
# manipulate the best-fit index. This can have multiple advantages if one
# wants to include additional information. But it should always be used
# carefully to not falsify or bias any results!
#
# Note how the :math:`\Delta`\ TS of the dataset with the prior set is
# larger (:math:`-53.91`) than the one without prior (:math:`-55.03`)
# since the index is not fitted to the underlying true value. If the
# Gaussian priors mean were the true value of :math:`2.3`, the index would
# be fitted correctly, and the :math:`\Delta`\ TS values would be the
# same.
#

######################################################################
# Implementing a custom prior
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# To add a use case specific prior one has to create a prior subclass
# containing:
#
# -  a `tag` and a `_type` used for the serialization
# -  an instantiation of each `PriorParameter` with their unit and default values
# -  the` evaluate` function where the mathematical expression for the prior is defined.
# -  if the prior has to be used for sampling it should also define a `_inverse_cdf` function
#    (in that case the pdf has to integrate to unity).
#
# As an example for a custom prior a Jeffrey prior for a scale parameter is chosen.
# The only parameter is ``sigma`` and the evaluation method return the squared inverse of ``sigma``.


from gammapy.modeling import PriorParameter
from gammapy.modeling.models import Model, Prior


class MyCustomPrior(Prior):
    """Custom Prior.


    Parameters
    ----------
    min : float
        Minimum value.
        Default is -inf.
    max : float
        Maximum value.
        Default is inf.
    """

    tag = ["MyCustomPrior"]
    sigma = PriorParameter(name="sigma", value=1, unit="")
    _type = "prior"

    @staticmethod
    def evaluate(value, sigma):
        """Evaluate the custom prior."""
        return value / sigma**2


# The custom prior is added to the PRIOR_REGISTRY so that it can be serialised.

from gammapy.modeling.models import PRIOR_REGISTRY

PRIOR_REGISTRY.append(MyCustomPrior)

# The custom prior is set on the index of a powerlaw spectral model and is evaluated.
customprior = MyCustomPrior(sigma=0.5)
pwl = PowerLawSpectralModel()
pwl.parameters["index"].prior = customprior
customprior(pwl.parameters["index"])

# The power law spectral model can be written into a dictionary.
# If a model is read in from this dictionary, the custom prior is still set on the index.

print(pwl.to_dict())
model_read = Model.from_dict(pwl.to_dict())
model_read.parameters.prior


######################################################################
# Example 2: Encouraging Positive Amplitude Values
# ------------------------------------------------
#
# This second example is a form of penalized likelihood.
# Let's asssume we want to encourage the amplitude to have
# positive, i.e. physical, values. Instead of setting hard bounds, we can
# also set a uniform penalisation, which prefers positive values to negatives.
# For this use case we are going to define a `UniformPenalty`.


class UniformPenalty(Prior):
    """Uniform Penalty.

    Returns 0 if the parameter value is in ]min, max[.
    Returns the penalty value, if otherwise.

    Parameters
    ----------
    min : float, optional
        Minimum value.
        Default is -`~numpy.inf`.
    max : float, optional
        Maximum value.
        Default is `~numpy.inf`.
    penalty : float, optional
        Penalization constant.
        Default is 1.
    """

    tag = ["UniformPenalty"]
    _type = "prior"
    min = PriorParameter(name="min", value=-np.inf, unit="")
    max = PriorParameter(name="max", value=np.inf, unit="")
    penalty = PriorParameter(name="penalty", value=1, unit="")

    @staticmethod
    def evaluate(value, min, max, penalty):
        """Evaluate the uniform prior."""
        if min < value < max:
            return 0.0
        else:
            return penalty

    @property
    def _inverse_cdf(self):
        raise ValueError(
            "UniformPenalty is not a distribution it cannot be used for sampling"
        )


######################################################################
# We set the amplitude of the power-law used to simulate the source to a very
# small value. Together with statistical fluctuations, this could result in some
# negative amplitude best-fit values.
#

model_weak = SkyModel(
    PowerLawSpectralModel(
        amplitude=1e-13 / u.cm**2 / u.s / u.TeV,
    ),
    name="weak-model",
)
model_weak_prior = model_weak.copy(name="weak-model-prior")
uniform_penalty = UniformPenalty(min=0, max=np.inf, penalty=2)
model_weak_prior.parameters["amplitude"].prior = uniform_penalty


######################################################################
# We set the minimum value to zero and the maximum value
# to infinity. Outside of ]min, max[ the prior statistic will be penalized by
# a contant value of 2 as given by the penalty parameter.
#

uni_prior_stat_sums = []
with model_weak_prior.parameters.restore_status():
    a_scan = np.linspace(-1, 1, 100)
    for a in a_scan:
        model_weak_prior.parameters["amplitude"].value = a
        uni_prior_stat_sums.append(model_weak_prior.parameters.prior_stat_sum())

plt.plot(
    a_scan,
    uni_prior_stat_sums,
    color="tab:orange",
    linestyle="dashed",
    label=f"Uniform Prior\n $min={uniform_penalty.min.value}$, penalty={uniform_penalty.penalty.value}",
)
plt.xlabel("Amplitude Value [1 / (TeV s cm2)]")
plt.ylabel("Prior")
plt.legend()
plt.xlim([-1, 1])
plt.show()

######################################################################
# Fitting Multiple Datasets with and without the Prior
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#
# To showcase how the uniform prior affects the fit results, :math:`100`
# datasets are created and fitted without and with the prior
#

results, results_prior = [], []
N = 100
dataset2 = dataset.copy()
for n in range(N):
    # simulating the dataset
    dataset2.models = model_weak.copy()
    dataset2.fake(random_state=42 + n)

    dataset2_prior = dataset2.copy()
    dataset2_prior.models = model_weak_prior.copy()
    # fitting without the prior
    fit = Fit()
    result = fit.optimize(dataset2)
    results.append(
        {
            "index": result.parameters["index"].value,
            "amplitude": result.parameters["amplitude"].value,
        }
    )
    # fitting with the prior
    fit_prior = Fit()
    result = fit_prior.optimize(dataset2_prior)
    results_prior.append(
        {
            "index": result.parameters["index"].value,
            "amplitude": result.parameters["amplitude"].value,
        }
    )


fig, axs = plt.subplots(1, 2, figsize=(7, 4))
for i, parname in enumerate(["index", "amplitude"]):
    par = np.array([_[parname] for _ in results])
    c, bins, _ = axs[i].hist(par, bins=20, alpha=0.5, label="Without Prior")
    par = np.array([_[parname] for _ in results_prior])
    axs[i].hist(par, bins=bins, alpha=0.5, color="tab:green", label="With Prior")
    axs[i].axvline(x=model_weak.parameters[parname].value, color="red")
    axs[i].set_xlabel(f"Reconstructed spectral\n {parname}")
    axs[i].legend()
plt.tight_layout()
plt.show()


######################################################################
# The distribution of the best-fit amplitudes shows how less best-fit
# amplitudes have negative values. This also has an effect on the
# distribution of the best-fit indices. How exactly the distribution
# changes depends on the value assigned to the uniform penalty. The
# larger the penalty, the less negative amplitudes.
#
# Note that the model parameters uncertainties are, per default, computed
# symmetrical. This can lead to incorrect
# uncertainties, especially with asymmetrical priors like the previous
# uniform. Calculating the uncertainties from the profile likelihood
# is advised. For more details see the :doc:`/tutorials/details/fitting` tutorial.
#



"""Plot spin frequency of the pulsar with time."""

import numpy as np
from astropy.units import Quantity
import matplotlib.pyplot as plt
from gammapy.astro.source import Pulsar

t = Quantity(np.logspace(0, 6, 100), "yr")

pulsar = Pulsar(P_0=Quantity(0.01, "s"), B="1e12 G")

plt.plot(t.value, 1 / pulsar.period(t).cgs.value)
plt.xlabel("time [years]")
plt.ylabel("frequency [1/s]")
plt.ylim(1e0, 1e2)
plt.loglog()
plt.show()



from typing import ClassVar, Literal, Optional, Union
import numpy as np
from astropy.coordinates import SkyCoord
from pydantic import ConfigDict
from gammapy.utils.metadata import (
    METADATA_FITS_KEYS,
    CreatorMetaData,
    MetaData,
    ObsInfoMetaData,
    PointingInfoMetaData,
)

__all__ = ["MapDatasetMetaData"]

MapDataset_METADATA_FITS_KEYS = {
    "MapDataset": {
        "event_types": "EVT_TYPE",
        "optional": "OPTIONAL",
    },
}

METADATA_FITS_KEYS.update(MapDataset_METADATA_FITS_KEYS)


class MapDatasetMetaData(MetaData):
    """Metadata containing information about the Dataset.

    Parameters
    ----------
    creation : `~gammapy.utils.CreatorMetaData`, optional
         The creation metadata.
    obs_info : list of `~gammapy.utils.ObsInfoMetaData`
        info about the observation.
    event_types : list of int or str
        Event types used in analysis.
    pointing: list of `~gammapy.utils.PointingInfoMetaData`
        Telescope pointing directions.
    optional : dict
        Additional optional metadata.
    """

    model_config = ConfigDict(coerce_numbers_to_str=True)

    _tag: ClassVar[Literal["MapDataset"]] = "MapDataset"
    creation: Optional[CreatorMetaData] = CreatorMetaData()
    obs_info: Optional[Union[ObsInfoMetaData, list[ObsInfoMetaData]]] = None
    pointing: Optional[Union[PointingInfoMetaData, list[PointingInfoMetaData]]] = None
    event_type: Optional[Union[str, list[str]]] = None
    optional: Optional[dict] = None

    def stack(self, other):
        kwargs = {}
        kwargs["creation"] = self.creation
        return self.__class__(**kwargs)

    @classmethod
    def _from_meta_table(cls, table):
        """Create MapDatasetMetaData from MapDataset.meta_table.

        Parameters
        ----------
        table: `~astropy.table.Table`
        """
        kwargs = {}
        kwargs["creation"] = CreatorMetaData()
        telescope = np.atleast_1d(table["TELESCOP"].data[0])
        obs_id = np.atleast_1d(table["OBS_ID"].data[0].astype(str))
        observation_mode = np.atleast_1d(table["OBS_MODE"].data[0])

        obs_info = []
        for i in range(len(obs_id)):
            obs_meta = ObsInfoMetaData(
                **{
                    "telescope": telescope[i],
                    "obs_id": obs_id[i],
                    "observation_mode": observation_mode[i],
                }
            )
            obs_info.append(obs_meta)
        kwargs["obs_info"] = obs_info

        pointing_radec, pointing_altaz = None, None
        if "RA_PNT" in table.colnames:
            pointing_radec = SkyCoord(
                ra=table["RA_PNT"].data[0], dec=table["DEC_PNT"].data[0], unit="deg"
            )
        if "ALT_PNT" in table.colnames:
            pointing_altaz = SkyCoord(
                alt=table["ALT_PNT"].data[0],
                az=table["AZ_PNT"].data[0],
                unit="deg",
                frame="altaz",
            )
        pointings = []
        for pra, paz in zip(
            np.atleast_1d(pointing_radec), np.atleast_1d(pointing_altaz)
        ):
            pointings.append(PointingInfoMetaData(radec_mean=pra, altaz_mean=paz))
        kwargs["pointing"] = pointings
        return cls(**kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Multiprocessing and multithreading setup."""

import importlib
import logging
from enum import Enum
from gammapy.utils.pbar import progress_bar

log = logging.getLogger(__name__)

__all__ = [
    "multiprocessing_manager",
    "run_multiprocessing",
    "BACKEND_DEFAULT",
    "N_JOBS_DEFAULT",
    "POOL_KWARGS_DEFAULT",
    "METHOD_DEFAULT",
    "METHOD_KWARGS_DEFAULT",
]


class ParallelBackendEnum(Enum):
    """Enum for parallel backend."""

    multiprocessing = "multiprocessing"
    ray = "ray"

    @classmethod
    def from_str(cls, value):
        """Get enum from string."""
        if value == "ray" and not is_ray_available():
            log.warning("Ray is not installed, falling back to multiprocessing backend")
            value = "multiprocessing"

        return cls(value)


class PoolMethodEnum(Enum):
    """Enum for pool method."""

    starmap = "starmap"
    apply_async = "apply_async"


BACKEND_DEFAULT = ParallelBackendEnum.multiprocessing
N_JOBS_DEFAULT = 1
ALLOW_CHILD_JOBS = False
POOL_KWARGS_DEFAULT = dict(processes=N_JOBS_DEFAULT)
METHOD_DEFAULT = PoolMethodEnum.starmap
METHOD_KWARGS_DEFAULT = {}


def get_multiprocessing():
    """Get multiprocessing module."""
    import multiprocessing

    return multiprocessing


def get_multiprocessing_ray():
    """Get multiprocessing module for ray backend."""
    import ray.util.multiprocessing as multiprocessing

    log.warning(
        "Gammapy support for parallelisation with ray is still a prototype and is not fully functional."
    )
    return multiprocessing


def is_ray_initialized():
    """Check if ray is initialized."""
    try:
        from ray import is_initialized

        return is_initialized()
    except ModuleNotFoundError:
        return False


def is_ray_available():
    """Check if ray is available."""
    try:
        importlib.import_module("ray")
        return True
    except ModuleNotFoundError:
        return False


class multiprocessing_manager:
    """Context manager to update the default configuration for multiprocessing.

    Only the default configuration will be modified, if class arguments like
    `n_jobs` and `parallel_backend` are set they will overwrite the default configuration.

    Parameters
    ----------
    backend : {'multiprocessing', 'ray'}
        Backend to use.
    pool_kwargs : dict
        Keyword arguments passed to the pool. The number of processes is limited
        to the number of physical CPUs.
    method : {'starmap', 'apply_async'}
        Pool method to use.
    method_kwargs : dict
        Keyword arguments passed to the method

    Examples
    --------
    ::

        import gammapy.utils.parallel as parallel
        from gammapy.estimators import FluxPointsEstimator

        fpe = FluxPointsEstimator(energy_edges=[1, 3, 10] * u.TeV)

        with parallel.multiprocessing_manager(
                backend="multiprocessing",
                pool_kwargs=dict(processes=2),
            ):
            fpe.run(datasets)
    """

    def __init__(self, backend=None, pool_kwargs=None, method=None, method_kwargs=None):
        global \
            BACKEND_DEFAULT, \
            POOL_KWARGS_DEFAULT, \
            METHOD_DEFAULT, \
            METHOD_KWARGS_DEFAULT, \
            N_JOBS_DEFAULT
        self._backend = BACKEND_DEFAULT
        self._pool_kwargs = POOL_KWARGS_DEFAULT
        self._method = METHOD_DEFAULT
        self._method_kwargs = METHOD_KWARGS_DEFAULT
        self._n_jobs = N_JOBS_DEFAULT
        if backend is not None:
            BACKEND_DEFAULT = ParallelBackendEnum.from_str(backend).value
        if pool_kwargs is not None:
            POOL_KWARGS_DEFAULT = pool_kwargs
            N_JOBS_DEFAULT = pool_kwargs.get("processes", N_JOBS_DEFAULT)
        if method is not None:
            METHOD_DEFAULT = PoolMethodEnum(method).value
        if method_kwargs is not None:
            METHOD_KWARGS_DEFAULT = method_kwargs

    def __enter__(self):
        pass

    def __exit__(self, type, value, traceback):
        global \
            BACKEND_DEFAULT, \
            POOL_KWARGS_DEFAULT, \
            METHOD_DEFAULT, \
            METHOD_KWARGS_DEFAULT, \
            N_JOBS_DEFAULT
        BACKEND_DEFAULT = self._backend
        POOL_KWARGS_DEFAULT = self._pool_kwargs
        METHOD_DEFAULT = self._method
        METHOD_KWARGS_DEFAULT = self._method_kwargs
        N_JOBS_DEFAULT = self._n_jobs


class ParallelMixin:
    """Mixin class to handle parallel processing."""

    _n_child_jobs = 1

    @property
    def n_jobs(self):
        """Number of jobs as an integer."""
        # TODO: this is somewhat unusual behaviour. It deviates from a normal default value handling
        if self._n_jobs is None:
            return N_JOBS_DEFAULT

        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, value):
        """Number of jobs setter as an integer."""
        if not isinstance(value, (int, type(None))):
            raise ValueError(
                f"Invalid type: {value!r}, and integer or None is expected."
            )

        self._n_jobs = value
        if ALLOW_CHILD_JOBS:
            self._n_child_jobs = value

    def _update_child_jobs(self):
        """needed because we can update only in the main process
        otherwise global ALLOW_CHILD_JOBS has default value"""
        if ALLOW_CHILD_JOBS:
            self._n_child_jobs = self.n_jobs
        else:
            self._n_child_jobs = 1

    @property
    def _get_n_child_jobs(self):
        """Number of allowed child jobs as an integer."""
        return self._n_child_jobs

    @property
    def parallel_backend(self):
        """Parallel backend as a string."""
        if self._parallel_backend is None:
            return BACKEND_DEFAULT

        return self._parallel_backend

    @parallel_backend.setter
    def parallel_backend(self, value):
        """Parallel backend setter (str)"""
        if value is None:
            self._parallel_backend = None
        else:
            self._parallel_backend = ParallelBackendEnum.from_str(value).value


def run_multiprocessing(
    func,
    inputs,
    backend=None,
    pool_kwargs=None,
    method=None,
    method_kwargs=None,
    task_name="",
):
    """Run function in a loop or in Parallel.

    Notes
    -----
    The progress bar can be displayed for this function.

    Parameters
    ----------
    func : function
        Function to run.
    inputs : list
        List of arguments to pass to the function.
    backend : {'multiprocessing', 'ray'}, optional
        Backend to use. Default is None.
    pool_kwargs : dict, optional
        Keyword arguments passed to the pool. The number of processes is limited
        to the number of physical CPUs. Default is None.
    method : {'starmap', 'apply_async'}
        Pool method to use. Default is "starmap".
    method_kwargs : dict, optional
        Keyword arguments passed to the method. Default is None.
    task_name : str, optional
        Name of the task to display in the progress bar. Default is "".
    """
    if backend is None:
        backend = BACKEND_DEFAULT

    if method is None:
        method = METHOD_DEFAULT

    if method_kwargs is None:
        method_kwargs = METHOD_KWARGS_DEFAULT

    if pool_kwargs is None:
        pool_kwargs = POOL_KWARGS_DEFAULT

    try:
        method_enum = PoolMethodEnum(method)
    except ValueError as e:
        raise ValueError(f"Invalid method: {method}") from e

    processes = pool_kwargs.get("processes", N_JOBS_DEFAULT)

    backend = ParallelBackendEnum.from_str(backend)
    multiprocessing = PARALLEL_BACKEND_MODULES[backend]()

    if backend == ParallelBackendEnum.multiprocessing:
        cpu_count = multiprocessing.cpu_count()

        if processes > cpu_count:
            log.info(f"Limiting number of processes from {processes} to {cpu_count}")
            processes = cpu_count

        if multiprocessing.current_process().name != "MainProcess":
            # with multiprocessing subprocesses cannot have childs (but possible with ray)
            processes = 1

    if processes == 1:
        return run_loop(
            func=func, inputs=inputs, method_kwargs=method_kwargs, task_name=task_name
        )

    if backend == ParallelBackendEnum.ray:
        address = "auto" if is_ray_initialized() else None
        pool_kwargs.setdefault("ray_address", address)

    log.info(f"Using {processes} processes to compute {task_name}")

    with multiprocessing.Pool(**pool_kwargs) as pool:
        pool_func = POOL_METHODS[method_enum]
        results = pool_func(
            pool=pool,
            func=func,
            inputs=inputs,
            method_kwargs=method_kwargs,
            task_name=task_name,
        )

    return results


def run_loop(func, inputs, method_kwargs=None, task_name=""):
    """Loop over inputs and run function."""
    results = []

    callback = method_kwargs.get("callback", None)

    for arguments in progress_bar(inputs, desc=task_name):
        result = func(*arguments)

        if callback is not None:
            result = callback(result)

        results.append(result)

    return results


def run_pool_star_map(pool, func, inputs, method_kwargs=None, task_name=""):
    """Run function in parallel."""
    return pool.starmap(func, progress_bar(inputs, desc=task_name), **method_kwargs)


def run_pool_async(pool, func, inputs, method_kwargs=None, task_name=""):
    """Run function in parallel async."""
    results = []

    for arguments in progress_bar(inputs, desc=task_name):
        result = pool.apply_async(func, arguments, **method_kwargs)
        results.append(result)
    # wait async run is done
    [result.wait() for result in results]
    return results


POOL_METHODS = {
    PoolMethodEnum.starmap: run_pool_star_map,
    PoolMethodEnum.apply_async: run_pool_async,
}

PARALLEL_BACKEND_MODULES = {
    ParallelBackendEnum.multiprocessing: get_multiprocessing,
    ParallelBackendEnum.ray: get_multiprocessing_ray,
}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.modeling.models import create_cosmic_ray_spectral_model

COSMIC_RAY_SPECTRA = [
    {"name": "proton", "dnde": 1.856522e-01, "flux": 7.096247e-01, "index": 2.70},
    {"name": "N", "dnde": 1.449504e-01, "flux": 5.509215e-01, "index": 2.64},
    {"name": "Si", "dnde": 5.646618e-02, "flux": 2.149887e-01, "index": 2.66},
    {"name": "Fe", "dnde": 2.720231e-02, "flux": 1.03305e-01, "index": 2.63},
    {"name": "electron", "dnde": 1.013671e-04, "flux": 4.691975e-04, "index": 3.428318},
]


@pytest.mark.parametrize("spec", COSMIC_RAY_SPECTRA, ids=lambda _: _["name"])
def test_cosmic_ray_spectrum(spec):
    cr_spectrum = create_cosmic_ray_spectral_model(particle=spec["name"])

    dnde = cr_spectrum(2 * u.TeV)
    assert_allclose(dnde.value, spec["dnde"], rtol=1e-3)
    assert dnde.unit == "m-2 s-1 TeV-1"

    flux = cr_spectrum.integral(1 * u.TeV, 1e3 * u.TeV)
    assert_allclose(flux.value, spec["flux"], rtol=1e-3)
    assert flux.unit == "m-2 s-1"

    index = cr_spectrum.spectral_index(2 * u.TeV)
    assert_allclose(index.value, spec["index"], rtol=1e-3)
    assert index.unit == ""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.table import Column, Table
from astropy.time import Time
from gammapy.estimators import FluxPoints
from gammapy.stats.variability import (
    TimmerKonig_lightcurve_simulator,
    compute_chisq,
    compute_flux_doubling,
    compute_fpp,
    compute_fvar,
    structure_function,
    discrete_correlation,
)
from gammapy.utils.testing import assert_quantity_allclose


@pytest.fixture(scope="session")
def lc_table():
    meta = dict(TIMESYS="utc")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-10"]).mjd, "time_max"),
            Column([1e-11, 3e-11], "flux", unit="cm-2 s-1"),
            Column([0.1e-11, 0.3e-11], "flux_err", unit="cm-2 s-1"),
            Column([np.nan, 3.6e-11], "flux_ul", unit="cm-2 s-1"),
            Column([False, True], "is_ul"),
        ],
    )

    return table


@pytest.fixture(scope="session")
def lc():
    meta = dict(TIMESYS="utc", SED_TYPE="flux")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-01", "2010-01-03", "2010-01-07"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-07", "2010-01-10"]).mjd, "time_max"),
            Column([[1.0, 2.0], [1.0, 2.0], [1.0, 2.0]], "e_min", unit="TeV"),
            Column([[2.0, 5.0], [2.0, 5.0], [2.0, 5.0]], "e_max", unit="TeV"),
            Column(
                [[1e-11, 4e-12], [3e-11, np.nan], [1e-11, 1e-12]],
                "flux",
                unit="cm-2 s-1",
            ),
            Column(
                [[0.1e-11, 0.4e-12], [0.3e-11, np.nan], [0.1e-11, 0.1e-12]],
                "flux_err",
                unit="cm-2 s-1",
            ),
            Column(
                [[np.nan, np.nan], [3.6e-11, 1e-11], [1e-11, 1e-12]],
                "flux_ul",
                unit="cm-2 s-1",
            ),
            Column([[False, False], [True, True], [True, True]], "is_ul"),
            Column([[True, True], [True, True], [True, True]], "success"),
        ],
    )

    return FluxPoints.from_table(table=table, format="lightcurve")


def test_lightcurve_fvar():
    flux = np.array([[1e-11, 4e-12], [3e-11, np.nan], [1e-11, 1e-12]])
    flux_err = np.array([[0.1e-11, 0.4e-12], [0.3e-11, np.nan], [0.1e-11, 0.1e-12]])

    time_id = 0

    fvar, fvar_err = compute_fvar(flux, flux_err, axis=time_id)

    assert_allclose(fvar, [0.68322763, 0.84047606])
    assert_allclose(fvar_err, [0.06679978, 0.08285806])


def test_lightcurve_fpp():
    flux = np.array([[1e-11, 4e-12], [3e-11, np.nan], [1e-11, 1e-12]])
    flux_err = np.array([[0.1e-11, 0.4e-12], [0.3e-11, np.nan], [0.1e-11, 0.1e-12]])

    fpp, fpp_err = compute_fpp(flux, flux_err)

    assert_allclose(fpp, [1.19448734, 0.11661904])
    assert_allclose(fpp_err, [0.06648574, 0.10099505])

    flux_1d = np.array([1e-11, 3e-11, 1e-11])
    flux_err_1d = np.array([0.1e-11, 0.3e-11, 0.1e-11])

    fpp_1d, fpp_err_1d = compute_fpp(flux_1d, flux_err_1d)

    assert_allclose(fpp_1d, [1.19448734])
    assert_allclose(fpp_err_1d, [0.06648574])


def test_lightcurve_chisq(lc_table):
    flux = lc_table["flux"].astype("float64")
    chi2, pval = compute_chisq(flux)
    assert_quantity_allclose(chi2, 1e-11)
    assert_quantity_allclose(pval, 0.999997476867478)


def test_lightcurve_flux_doubling():
    flux = np.array(
        [
            [1e-11, 4e-12],
            [3e-11, np.nan],
            [1e-11, 1e-12],
            [0.8e-11, 0.8e-12],
            [1e-11, 1e-12],
        ]
    )
    flux_err = np.array(
        [
            [0.1e-11, 0.4e-12],
            [0.3e-11, np.nan],
            [0.1e-11, 0.1e-12],
            [0.08e-11, 0.8e-12],
            [0.1e-11, 0.1e-12],
        ]
    )
    time = (
        np.array(
            [6.31157019e08, 6.31160619e08, 6.31164219e08, 6.31171419e08, 6.31178419e08]
        )
        * u.s
    )
    time_id = 0

    dtime_dict = compute_flux_doubling(flux, flux_err, time, axis=time_id)

    dtime = dtime_dict["doubling"]
    dtime_err = dtime_dict["doubling_err"]
    assert_allclose(
        dtime,
        [2271.34711286, 21743.98603654] * u.s,
    )
    assert_allclose(dtime_err, [425.92375713, 242.80234065] * u.s)


def test_tk_function():
    time_series, time_axis = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-3), 20, 1 * u.s
    )

    def temp(x, norm, index):
        return norm * x ** (-index)

    params = {"norm": 1.5, "index": 3}

    time_series2, time_axis2 = TimmerKonig_lightcurve_simulator(
        temp, 15, 1 * u.h, power_spectrum_params=params
    )

    assert len(time_series) == 20
    assert isinstance(time_axis, u.Quantity)
    assert time_axis.unit == u.s
    assert len(time_series2) == 15


def test_tk_nchunks():
    time_series, time_axis = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-1.5), 21, 1 * u.s, nchunks=100
    )
    time_series2, time_axis2 = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-1.5), 21, 1 * u.s, nchunks=20
    )

    with pytest.raises(TypeError):
        TimmerKonig_lightcurve_simulator(lambda x: x ** (-3), 20, 1 * u.s, nchunks=0.5)

    with pytest.raises(TypeError):
        TimmerKonig_lightcurve_simulator(lambda x: x ** (-3), 20.5, 1 * u.s)

    assert len(time_series) == len(time_series2) == 21


def test_tk_mean():
    time_series, time_axis = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-1.5), 2000, 1 * u.s, mean=2.5, std=0.5
    )

    time_series2, time_axis2 = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-1.5),
        2000,
        1 * u.s,
        mean=1e-7 * (u.cm**-2),
        std=5e-8 * (u.cm**-2),
    )

    time_series3, time_axis3 = TimmerKonig_lightcurve_simulator(
        lambda x: x ** (-1.5), 2000, 1 * u.s, mean=2.5, std=0.5, poisson=True
    )

    with pytest.raises(Warning):
        TimmerKonig_lightcurve_simulator(
            lambda x: x ** (-3), 20, 1 * u.s, mean=0.5, poisson=True
        )

    assert_allclose(time_series.mean(), 2.5)
    assert_allclose(time_series.std(), 0.5)
    assert_allclose(time_series2.mean(), 1e-7 * (u.cm**-2))
    assert_allclose(time_series2.std(), 5e-8 * (u.cm**-2))
    assert_allclose(time_series3.mean(), 2.5, rtol=0.1)
    assert_allclose(time_series3.std(), 1, rtol=1)


def test_structure_function():
    flux = np.array(
        [
            [1e-11, 4e-12],
            [3e-11, 2.5e-12],
            [1e-11, 1e-12],
            [0.8e-11, 0.8e-12],
            [1e-11, 1e-12],
        ]
    )
    flux_err = np.array(
        [
            [0.1e-11, 0.4e-12],
            [0.3e-11, 0.2e-12],
            [0.1e-11, 0.1e-12],
            [0.08e-11, 0.08e-12],
            [0.1e-11, 0.1e-12],
        ]
    )
    time = (
        np.array(
            [6.31157019e08, 6.31160619e08, 6.31164219e08, 6.31171419e08, 6.31178419e08]
        )
        * u.s
    )

    sf, distances = structure_function(flux, flux_err, time)

    assert_allclose(
        sf,
        [
            [4.00000000e-22, 2.25000000e-24],
            [4.00000000e-24, 4.00000000e-26],
            [2.00000000e-24, 4.52000000e-24],
            [4.84000000e-22, 2.89000000e-24],
            [0.00000000e00, 0.00000000e00],
            [4.00000000e-24, 1.02400000e-23],
            [4.00000000e-22, 2.25000000e-24],
            [0.00000000e00, 9.00000000e-24],
        ],
    )

    assert_allclose(
        distances,
        [3600.0, 7000.0, 7200.0, 10800.0, 14200.0, 14400.0, 17800.0, 21400.0] * u.s,
    )


def test_dcf():
    flux = np.array(
        [
            [1e-11, 4e-12],
            [3e-11, 2.5e-12],
            [1e-11, 1e-12],
            [0.8e-11, 0.8e-12],
            [1e-11, 1e-12],
        ]
    )
    flux2 = np.array(
        [
            [1e-11, 3.5e-12],
            [4e-11, 1.5e-12],
            [1.7e-11, 2e-12],
            [0.4e-11, 1.8e-12],
            [1e-11, 1.5e-12],
            [1.1e-11, 0.7e-12],
        ]
    )

    flux_err = np.array(
        [
            [0.1e-11, 0.4e-12],
            [0.3e-11, 0.2e-12],
            [0.1e-11, 0.1e-12],
            [0.08e-11, 0.08e-12],
            [0.1e-11, 0.1e-12],
        ]
    )
    flux_err2 = np.array(
        [
            [0.2e-11, 0.24e-12],
            [0.4e-11, 0.21e-12],
            [0.07e-11, 0.14e-12],
            [0.12e-11, 0.18e-12],
            [0.1e-11, 0.13e-12],
            [0.4e-11, 0.2e-12],
        ]
    )

    time = (
        np.array(
            [6.31157019e08, 6.31160619e08, 6.31164219e08, 6.31171419e08, 6.31178419e08]
        )
        * u.s
    )
    time2 = (
        np.array(
            [
                6.31156919e08,
                6.31160519e08,
                6.31163819e08,
                6.31171419e08,
                6.31178219e08,
                6.31178819e08,
            ]
        )
        * u.s
    )

    bins, dcf, dcf_err = discrete_correlation(
        flux, flux_err, flux2, flux_err2, time, time2, tau=1.5e4 * u.s
    )

    assert_allclose(
        dcf,
        [
            [-0.332361, -1.009638],
            [-0.065639, 0.313054],
            [0.215329, 0.133132],
            [-0.373906, -0.56788],
        ],
        rtol=1e-5,
    )

    assert_allclose(
        dcf_err,
        [
            [0.34881, 0.553113],
            [0.236548, 0.174538],
            [0.401313, 0.361757],
            [0.820525, 1.204655],
        ],
        rtol=1e-5,
    )

    assert_allclose(
        bins,
        u.Quantity([-22500.0, -7500.0, 7500.0, 22500.0], unit="s"),
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from gammapy.modeling import Parameter, Parameters
from gammapy.modeling.scipy import (
    confidence_scipy,
    optimize_scipy,
    stat_profile_ul_scipy,
)
from gammapy.estimators import FluxPoints
from gammapy.utils.testing import requires_data


class MyDataset:
    def __init__(self, parameters):
        self.parameters = parameters

    def fcn(self):
        x, y, z = [p.value for p in self.parameters]
        x_opt, y_opt, z_opt = 2, 3e5, 4e-5
        x_err, y_err, z_err = 0.2, 3e4, 4e-6
        return (
            ((x - x_opt) / x_err) ** 2
            + ((y - y_opt) / y_err) ** 2
            + ((z - z_opt) / z_err) ** 2
        )


@pytest.fixture()
def pars():
    x = Parameter("x", 2.1)
    y = Parameter("y", 3.1e5, scale=1e5)
    z = Parameter("z", 4.1e-5, scale=1e-5)
    return Parameters([x, y, z])


@pytest.mark.parametrize("method", ["Nelder-Mead", "L-BFGS-B", "Powell", "BFGS"])
def test_scipy_basic(pars, method):
    ds = MyDataset(pars)
    factors, info, optimizer = optimize_scipy(
        function=ds.fcn, parameters=pars, method=method
    )

    assert info["success"]
    assert_allclose(ds.fcn(), 0, atol=1e-5)

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2, rtol=1e-3)
    assert_allclose(pars["y"].value, 3e5, rtol=1e-3)
    assert_allclose(pars["z"].value, 4e-5, rtol=2e-2)


@pytest.mark.parametrize("method", ["Nelder-Mead", "L-BFGS-B", "Powell"])
def test_scipy_frozen(pars, method):
    ds = MyDataset(pars)
    pars["y"].frozen = True

    factors, info, optimizer = optimize_scipy(
        function=ds.fcn, parameters=pars, method=method
    )

    assert info["success"]

    assert_allclose(pars["x"].value, 2, rtol=1e-4)
    assert_allclose(pars["y"].value, 3.1e5)
    assert_allclose(pars["z"].value, 4.0e-5, rtol=1e-4)
    assert_allclose(ds.fcn(), 0.1111111, rtol=1e-5)


@pytest.mark.parametrize("method", ["L-BFGS-B"])
def test_scipy_limits(pars, method):
    ds = MyDataset(pars)
    pars["y"].min = 301000

    factors, info, minuit = optimize_scipy(
        function=ds.fcn, parameters=pars, method=method
    )

    assert info["success"]

    # Check the result in parameters is OK
    assert_allclose(pars["x"].value, 2, rtol=1e-2)
    assert_allclose(pars["y"].value, 301000, rtol=1e-3)


def test_scipy_confidence(pars):
    ds = MyDataset(pars)
    factors, info, _ = optimize_scipy(function=ds.fcn, parameters=pars)

    assert_allclose(ds.fcn(), 0, atol=1e-5)

    par = pars["x"]
    par.scan_min, par.scan_max = 0, 10

    result = confidence_scipy(function=ds.fcn, parameters=pars, parameter=par, sigma=1)

    assert result["success_errp"]
    assert result["success_errn"]

    assert_allclose(result["errp"], 0.2, rtol=1e-3)
    assert_allclose(result["errn"], 0.2, rtol=1e-3)


@requires_data()
def test_stat_profile_ul_scipy():
    # Test normal profile with a minima
    x = np.linspace(-5, 5, 7)
    y = x**2
    ul = stat_profile_ul_scipy(x, y)
    assert_allclose(ul, 2)

    ul = stat_profile_ul_scipy(x, y, interp_scale="lin")
    assert_allclose(ul, 1.9111111)

    # Test with real data
    flux_point = FluxPoints.read(
        "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
        format="lightcurve",
    )
    value_scan = flux_point.stat_scan.geom.axes["norm"].center
    stat_scan = np.sum(flux_point.stat_scan.data, axis=0).ravel()
    ul = stat_profile_ul_scipy(value_scan, stat_scan, delta_ts=4)
    assert_allclose(ul, 1.1123425)

    # Test a flat profile i.e. all minima
    x = np.linspace(0, 10, 10)
    y = np.ones_like(x) * 5

    with pytest.raises(
        ValueError,
        match="Statistic profile is flat therefore no best-fit value can be determined.",
    ):
        stat_profile_ul_scipy(x, y)

    # Test with NaN result
    x = np.linspace(0, 5, 20)
    y = (x - 2.5) ** 2 + 10

    with pytest.raises(
        RuntimeError, match="Failed to find upper limit: no valid root found."
    ):
        stat_profile_ul_scipy(x, y, delta_ts=20)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gammapy high level interface (workflow)."""
from gammapy.utils.registry import Registry
from .config import WorkflowConfig
from .core import Workflow
from .steps import (
    DataSelectionWorkflowStep,
    DatasetsWorkflowStep,
    ExcessMapWorkflowStep,
    FitWorkflowStep,
    FluxPointsWorkflowStep,
    LightCurveWorkflowStep,
    ObservationsWorkflowStep,
)

__all__ = [
    "Workflow",
    "WorkflowConfig",
]

WORKFLOW_STEP_REGISTRY = Registry(
    [
        DataSelectionWorkflowStep,
        ObservationsWorkflowStep,
        DatasetsWorkflowStep,
        ExcessMapWorkflowStep,
        FitWorkflowStep,
        FluxPointsWorkflowStep,
        LightCurveWorkflowStep,
    ]
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
import matplotlib.pyplot as plt

# taken from the matploltlib documentation
# https://matplotlib.org/3.1.0/gallery/images_contours_and_fields/image_annotated_heatmap.html#sphx-glr-gallery-images-contours-and-fields-image-annotated-heatmap-py

__all__ = [
    "annotate_heatmap",
    "plot_heatmap",
]


def plot_heatmap(
    data, row_labels, col_labels, ax=None, cbar_kw=None, cbarlabel="", **kwargs
):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data : `~numpy.ndarray`
        Data array.
    row_labels : list or `~numpy.ndarray`
        List or array of labels for the rows.
    col_labels : list or `~numpy.ndarray`
        List or array of labels for the columns.
    ax : `matplotlib.axes.Axes`, optional
        Axis instance to which the heatmap is plotted. Default is None. If None, the current one is used.
    cbar_kw : dict, optional
        A dictionary with arguments to `matplotlib.Figure.colorbar`. Default is None.
    cbarlabel : str, optional
        The label for the color bar. Default is "".
    **kwargs : dict, optional
        Other keyword arguments forwarded to `matplotlib.axes.Axes.imshow`.
    """
    if ax is None:
        ax = plt.gca()

    if cbar_kw is None:
        cbar_kw = {}

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False, labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=-30, ha="right", rotation_mode="anchor")

    # Turn spines off and create white grid.
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(np.arange(data.shape[1] + 1) - 0.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0] + 1) - 0.5, minor=True)
    ax.grid(which="minor", color="w", linestyle="-", linewidth=1.5)
    ax.tick_params(which="minor", bottom=False, left=False)

    return im, cbar


def annotate_heatmap(
    im,
    data=None,
    valfmt="{x:.2f}",
    textcolors=("black", "white"),
    threshold=None,
    **textkw,
):
    """
    A function to annotate a heatmap.

    Parameters
    ----------
    im
        The AxesImage to be labeled.
    data : `~numpy.ndarray`, optional
        Data used to annotate. Default is None. If None, the image's data is used.
    valfmt : str format or `matplotlib.ticker.Formatter`, optional
        The format of the annotations inside the heatmap. This should either
        use the string format method, e.g. "$ {x:.2f}"
        or be a `matplotlib.ticker.Formatter` instance. Default is "{x:.2f}".
    textcolors : list or `~numpy.ndarray`, optional
        Two color specifications.  The first is used for
        values below a threshold, the second for those above. Default is ["black", "white"].
    threshold : float, optional
        Value in data units according to which the colors from textcolors are
        applied. Default is None. If None the middle of the colormap is used as
        separation.
    **kwargs : dict, optional
        Other keyword arguments forwarded to each call to `text` used to create
        the text labels.
    """
    import matplotlib

    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    # Normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max()) / 2.0

    # Set default alignment to center, but allow it to be
    # overwritten by textkw.
    kw = dict(horizontalalignment="center", verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import astropy.units as u
from astropy.table import Table
import numpy as np
from regions import PointSkyRegion
from gammapy.datasets import MapDatasetMetaData
from gammapy.irf import EDispKernelMap, PSFMap
from gammapy.data import Observation
from gammapy.maps import Map
from .core import Maker
from .utils import (
    make_counts_rad_max,
    make_edisp_kernel_map,
    make_edisp_map,
    make_map_background_irf,
    make_map_exposure_true_energy,
    make_psf_map,
)

__all__ = ["MapDatasetMaker"]

log = logging.getLogger(__name__)


class MapDatasetMaker(Maker):
    """Make binned maps for a single IACT observation.

    Parameters
    ----------
    selection : list of str, optional
        Select which maps to make, the available options are:
        'counts', 'exposure', 'background', 'psf', 'edisp'.
        By default, all maps are made.
    background_oversampling : int
        Background evaluation oversampling factor in energy.
    background_interp_missing_data : bool, optional
        Interpolate missing values in background 3d map.
        Default is True, have to be set to True for CTAO IRF.
    background_pad_offset : bool, optional
        Pad one bin in offset for 2d background map.
        This avoids extrapolation at edges and use the nearest value.
        Default is True.
    fov_rotation_step : `~astropy.units.Quantity`, optional
        Maximum error on the rotation angle between AltAz and RaDec frames during background evaluation.
        Used only when the Background IRF has an AltAz alignement.
        Default is 1.0 deg.

    Examples
    --------
    This example shows how to run the MapMaker for a single observation.

    >>> from gammapy.data import DataStore
    >>> from gammapy.datasets import MapDataset
    >>> from gammapy.maps import WcsGeom, MapAxis
    >>> from gammapy.makers import MapDatasetMaker

    >>> # Load an observation
    >>> data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    >>> obs = data_store.obs(23523)

    >>> # Prepare the geometry
    >>> energy_axis = MapAxis.from_energy_bounds(1.0, 10.0, 4, unit="TeV")
    >>> energy_axis_true = MapAxis.from_energy_bounds( 0.5, 20, 10, unit="TeV", name="energy_true")
    >>> geom = WcsGeom.create(
    ...        skydir=(83.633, 22.014),
    ...        binsz=0.02,
    ...        width=(2, 2),
    ...        frame="icrs",
    ...        proj="CAR",
    ...        axes=[energy_axis],
    ...    )

    >>> # Run the maker
    >>> empty = MapDataset.create(geom=geom, energy_axis_true=energy_axis_true, name="empty")
    >>> maker = MapDatasetMaker()
    >>> dataset = maker.run(empty, obs)
    >>> print(dataset)
    MapDataset
    ----------
    <BLANKLINE>
      Name                            : empty
    <BLANKLINE>
      Total counts                    : 787
      Total background counts         : 684.52
      Total excess counts             : 102.48
    <BLANKLINE>
      Predicted counts                : 684.52
      Predicted background counts     : 684.52
      Predicted excess counts         : nan
    <BLANKLINE>
      Exposure min                    : 7.01e+07 m2 s
      Exposure max                    : 1.10e+09 m2 s
    <BLANKLINE>
      Number of total bins            : 40000
      Number of fit bins              : 40000
    <BLANKLINE>
      Fit statistic type              : cash
      Fit statistic value (-2 log(L)) : nan
    <BLANKLINE>
      Number of models                : 0
      Number of parameters            : 0
      Number of free parameters       : 0

    """

    tag = "MapDatasetMaker"
    available_selection = ["counts", "exposure", "background", "psf", "edisp"]

    def __init__(
        self,
        selection=None,
        background_oversampling=None,
        background_interp_missing_data=True,
        background_pad_offset=True,
        fov_rotation_step=1.0 * u.deg,
    ):
        self.background_oversampling = background_oversampling
        self.background_interp_missing_data = background_interp_missing_data
        self.background_pad_offset = background_pad_offset
        self.fov_rotation_step = fov_rotation_step
        if selection is None:
            selection = self.available_selection

        selection = set(selection)

        if not selection.issubset(self.available_selection):
            difference = selection.difference(self.available_selection)
            raise ValueError(f"{difference} is not a valid method.")

        self.selection = selection

    @staticmethod
    def make_counts(geom, observation):
        """Make counts map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference map geometry.
        observation : `~gammapy.data.Observation`
            Observation container.

        Returns
        -------
        counts : `~gammapy.maps.Map`
            Counts map.
        """
        if geom.is_region and isinstance(geom.region, PointSkyRegion):
            counts = make_counts_rad_max(geom, observation.rad_max, observation.events)
        else:
            counts = Map.from_geom(geom)
            counts.fill_events(observation.events)
        return counts

    @staticmethod
    def make_exposure(geom, observation, use_region_center=True):
        """Make exposure map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference map geometry.
        observation : `~gammapy.data.Observation`
            Observation container.
        use_region_center : bool, optional
            For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
            If False, average over the whole region.
            Default is True.

        Returns
        -------
        exposure : `~gammapy.maps.Map`
            Exposure map.
        """
        if getattr(observation, "exposure", None):
            return observation.exposure.interp_to_geom(
                geom=geom,
            )
        with u.add_enabled_units([u.def_unit("transit", u.sday)]):
            if isinstance(observation.aeff, Map):
                aeff = observation.aeff
                if aeff.unit.is_equivalent(u.Unit("m2 s")):
                    factor = 1.0
                elif aeff.unit.is_equivalent(u.Unit("m2 s transit-1")):
                    if u.Unit("transit") in aeff.unit.bases:
                        # TODO : GTI = number of transits is valid only
                        # if the exposure is flat in right ascension.
                        # It's fine for now as irfs are created like this for HAWC/SWGO
                        # but could change in future.
                        factor = observation.gti.time_sum.to("transit")
                    else:
                        factor = observation.gti.time_sum
                else:
                    raise u.UnitsError(
                        f"Effective area unit {observation.aeff.unit} is not supported"
                    )
                return observation.aeff.interp_to_geom(geom=geom) * factor

        return make_map_exposure_true_energy(
            pointing=observation.get_pointing_icrs(observation.tmid),
            livetime=observation.observation_live_time_duration,
            aeff=observation.aeff,
            geom=geom,
            use_region_center=use_region_center,
        )

    @staticmethod
    def make_exposure_irf(geom, observation, use_region_center=True):
        """Make exposure map with IRF geometry.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference geometry.
        observation : `~gammapy.data.Observation`
            Observation container.
        use_region_center : bool, optional
            For geom as a `~gammapy.maps.RegionGeom`. If True, consider the values at the region center.
            If False, average over the whole region.
            Default is True.

        Returns
        -------
        exposure : `~gammapy.maps.Map`
            Exposure map.
        """
        return make_map_exposure_true_energy(
            pointing=observation.get_pointing_icrs(observation.tmid),
            livetime=observation.observation_live_time_duration,
            aeff=observation.aeff,
            geom=geom,
            use_region_center=use_region_center,
        )

    def make_background(self, geom, observation):
        """Make background map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference geometry.
        observation : `~gammapy.data.Observation`
            Observation container.

        Returns
        -------
        background : `~gammapy.maps.Map`
            Background map.
        """
        bkg = observation.bkg
        with u.add_enabled_units([u.def_unit("transit", u.sday)]):
            if isinstance(bkg, Map):
                if not bkg.unit or bkg.unit.is_equivalent(u.Unit("")):
                    return bkg.interp_to_geom(geom=geom, preserve_counts=True)
                elif bkg.unit.is_equivalent(
                    u.Unit("TeV-1 sr-1")
                ) or bkg.unit.is_equivalent(u.Unit("TeV-1 sr-1 transit-1")):
                    bkg = bkg.interp_to_geom(geom, preserve_counts=False)
                    bkg.quantity *= bkg.geom.to_image().solid_angle()[None, :, :]
                    # multipling by energy bins after take less memory than using bin_volume
                    bkg.quantity *= np.diff(bkg.geom.axes["energy"].edges)[
                        :, None, None
                    ]
                    # TODO : would be better to have proper integration method
                    if bkg.unit.is_equivalent(u.Unit("transit-1")):
                        if u.Unit("transit") in bkg.unit.bases:
                            bkg.quantity *= observation.gti.time_sum.to("transit")
                        else:
                            bkg.quantity *= observation.gti.time_sum
                    return bkg
                else:
                    raise u.UnitsError(
                        f"Background unit {observation.bkg.unit} is not supported"
                    )

        use_region_center = getattr(self, "use_region_center", True)

        if self.background_interp_missing_data:
            bkg.interp_missing_data(axis_name="energy")

        if self.background_pad_offset and bkg.has_offset_axis:
            bkg = bkg.pad(1, mode="edge", axis_name="offset")

        return make_map_background_irf(
            pointing=observation.pointing,
            ontime=observation.observation_time_duration,
            bkg=bkg,
            geom=geom,
            time_start=observation.tstart,
            fov_rotation_step=self.fov_rotation_step,
            oversampling=self.background_oversampling,
            use_region_center=use_region_center,
            location=observation.observatory_earth_location,
        )

    def make_edisp(self, geom, observation):
        """Make energy dispersion map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference geometry.
        observation : `~gammapy.data.Observation`
            Observation container.

        Returns
        -------
        edisp : `~gammapy.irf.EDispMap`
            Energy dispersion map.
        """
        exposure = self.make_exposure_irf(geom.squash(axis_name="migra"), observation)

        use_region_center = getattr(self, "use_region_center", True)

        return make_edisp_map(
            edisp=observation.edisp,
            pointing=observation.get_pointing_icrs(observation.tmid),
            geom=geom,
            exposure_map=exposure,
            use_region_center=use_region_center,
        )

    def make_edisp_kernel(self, geom, observation):
        """Make energy dispersion kernel map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference geometry. Must contain "energy" and "energy_true" axes in that order.
        observation : `~gammapy.data.Observation`
            Observation container.

        Returns
        -------
        edisp : `~gammapy.irf.EDispKernelMap`
            Energy dispersion kernel map.
        """
        edisp = observation.edisp
        if isinstance(edisp, EDispKernelMap):
            if edisp.exposure_map:
                exposure = edisp.exposure_map.interp_to_geom(
                    geom.squash(axis_name="energy")
                )
            else:
                exposure = None
            interp_map = edisp.edisp_map.interp_to_geom(geom)
            return EDispKernelMap(edisp_kernel_map=interp_map, exposure_map=exposure)

        exposure = self.make_exposure_irf(geom.squash(axis_name="energy"), observation)

        use_region_center = getattr(self, "use_region_center", True)

        return make_edisp_kernel_map(
            edisp=observation.edisp,
            pointing=observation.get_pointing_icrs(observation.tmid),
            geom=geom,
            exposure_map=exposure,
            use_region_center=use_region_center,
        )

    def make_psf(self, geom, observation):
        """Make PSF map.

        Parameters
        ----------
        geom : `~gammapy.maps.Geom`
            Reference geometry.
        observation : `~gammapy.data.Observation`
            Observation container.

        Returns
        -------
        psf : `~gammapy.irf.PSFMap`
            PSF map.
        """
        psf = observation.psf

        if isinstance(psf, PSFMap):
            if psf.exposure_map:
                exposure_map = psf.exposure_map.interp_to_geom(
                    geom.squash(axis_name="rad")
                )
            else:
                exposure_map = None
            return psf.__class__(psf.psf_map.interp_to_geom(geom), exposure_map)

        exposure = self.make_exposure_irf(geom.squash(axis_name="rad"), observation)

        return make_psf_map(
            psf=psf,
            pointing=observation.get_pointing_icrs(observation.tmid),
            geom=geom,
            exposure_map=exposure,
        )

    @staticmethod
    def make_meta_table(observation):
        """Make information meta table.

        Parameters
        ----------
        observation : `~gammapy.data.Observation`
            Observation.

        Returns
        -------
        meta_table : `~astropy.table.Table`
            Meta table.
        """
        row = {}
        row["TELESCOP"] = observation.aeff.meta.get("TELESCOP", "Unknown")
        row["OBS_ID"] = observation.obs_id

        row.update(observation.pointing.to_fits_header())

        meta_table = Table([row])
        if "ALT_PNT" in meta_table.colnames:
            meta_table["ALT_PNT"].unit = u.deg
            meta_table["AZ_PNT"].unit = u.deg
        if "RA_PNT" in meta_table.colnames:
            meta_table["RA_PNT"].unit = u.deg
            meta_table["DEC_PNT"].unit = u.deg

        return meta_table

    @staticmethod
    def _make_metadata(table):
        return MapDatasetMetaData._from_meta_table(table)

    def run(self, dataset, observation):
        """Make map dataset.

        Parameters
        ----------
        dataset : `~gammapy.datasets.MapDataset`
            Reference dataset.
        observation : `~gammapy.data.Observation`
            Observation.

        Returns
        -------
        dataset : `~gammapy.datasets.MapDataset`
            Map dataset.
        """
        kwargs = {"gti": observation.gti}
        if isinstance(observation, Observation):
            kwargs["meta_table"] = self.make_meta_table(observation)
            kwargs["meta"] = self._make_metadata(kwargs["meta_table"])
        elif getattr(observation, "meta"):
            kwargs["meta"] = observation.meta

        mask_safe = Map.from_geom(dataset.counts.geom, dtype=bool)
        mask_safe.data[...] = True

        kwargs["mask_safe"] = mask_safe

        if "counts" in self.selection:
            counts = self.make_counts(dataset.counts.geom, observation)
        else:
            counts = Map.from_geom(dataset.counts.geom, data=0)
        kwargs["counts"] = counts

        if "exposure" in self.selection:
            exposure = self.make_exposure(dataset.exposure.geom, observation)
            kwargs["exposure"] = exposure

        if "background" in self.selection:
            kwargs["background"] = self.make_background(
                dataset.counts.geom, observation
            )

        if "psf" in self.selection:
            psf = self.make_psf(dataset.psf.psf_map.geom, observation)
            kwargs["psf"] = psf

        if "edisp" in self.selection:
            if dataset.edisp.edisp_map.geom.axes[0].name.upper() == "MIGRA":
                edisp = self.make_edisp(dataset.edisp.edisp_map.geom, observation)
            else:
                edisp = self.make_edisp_kernel(
                    dataset.edisp.edisp_map.geom, observation
                )

            kwargs["edisp"] = edisp

        return dataset.__class__(name=dataset.name, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities for hierarchical/agglomerative clustering."""
import numpy as np
import scipy.cluster.hierarchy as sch

__all__ = ["standard_scaler", "hierarchical_clustering"]


def standard_scaler(features):
    r"""Compute standardized features by removing the mean and scaling to unit variance.

    Calculated through:

    .. math::
        f_\text{scaled} = \frac{f-\text{mean}(f)}{\text{std}(f)} .

    Parameters
    ----------
    features : `~astropy.table.Table`
        Table containing the features.

    Returns
    -------
    scaled_features : `~astropy.table.Table`
        Table containing the scaled features (dimensionless).


    Examples
    --------
    Compute standardized features of a cluster observations based on their IRF quantities::

    >>> from gammapy.data.utils import get_irfs_features
    >>> from gammapy.data import DataStore
    >>> from gammapy.utils.cluster import standard_scaler
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u

    >>> data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    >>> obs_ids = data_store.obs_table["OBS_ID"][:3]
    >>> obs = data_store.get_observations(obs_ids)

    >>> position = SkyCoord(329.716 * u.deg, -30.225 * u.deg, frame="icrs")
    >>> names = ["edisp-res", "psf-radius"]
    >>> features_irfs = get_irfs_features(
    ...     obs,
    ...     energy_true="1 TeV",
    ...     position=position,
    ...     names=names
    ... )
    >>> scaled_features_irfs = standard_scaler(features_irfs)
    >>> print(scaled_features_irfs)
         edisp-res      obs_id      psf-radius
    ------------------- ------ --------------------
    -0.1379190199428797  20136 -0.18046952655570045
     1.2878662980210884  20137   1.3049664466089965
    -1.1499472780781963  20151  -1.1244969200533408
    """
    scaled_features = features.copy()
    for col in scaled_features.columns:
        if col not in ["obs_id", "dataset_name"]:
            data = scaled_features[col].data
            scaled_features[col] = (data - data.mean()) / data.std()
    return scaled_features


def hierarchical_clustering(features, linkage_kwargs=None, fcluster_kwargs=None):
    """Hierarchical clustering using given features.

    Parameters
    ----------
    features : `~astropy.table.Table`
        Table containing the features.
    linkage_kwargs : dict, optional
        Arguments forwarded to `scipy.cluster.hierarchy.linkage`.
        Default is None, which uses method="ward" and metric="euclidean".
    fcluster_kwargs : dict, optional
        Arguments forwarded to `scipy.cluster.hierarchy.fcluster`.
        Default is None, which uses criterion="maxclust" and t=3.


    Returns
    -------
    features : `~astropy.table.Table`
        Table containing the features and an extra column for the groups labels.


    Examples
    --------
    Cluster features into t=2 groups with a corresponding label for each group::

    >>> from gammapy.data.utils import get_irfs_features
    >>> from gammapy.data import DataStore
    >>> from gammapy.utils.cluster import standard_scaler, hierarchical_clustering
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u

    >>> data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    >>> obs_ids = data_store.obs_table["OBS_ID"][13:20]
    >>> obs = data_store.get_observations(obs_ids)

    >>> position = SkyCoord(329.716 * u.deg, -30.225 * u.deg, frame="icrs")
    >>> names = ["edisp-res", "psf-radius"]
    >>> features_irfs = get_irfs_features(
    ...     obs,
    ...     energy_true="1 TeV",
    ...     position=position,
    ...     names=names
    ... )
    >>> scaled_features_irfs = standard_scaler(features_irfs)

    >>> features = hierarchical_clustering(scaled_features_irfs, fcluster_kwargs={"t": 2})
    >>> print(features)
         edisp-res      obs_id      psf-radius     labels
    ------------------- ------ ------------------- ------
    -1.3020791585772495  20326 -1.2471938975366008      2
    -1.3319831545301117  20327 -1.4586649826004114      2
    -0.7763307219821931  20339 -0.6705024680435898      2
     0.9677107409819438  20343  0.9500979841335693      1
      0.820562952023891  20344  0.8160964882165554      1
     0.7771617763704126  20345  0.7718272408581743      1
     0.8449575657133206  20346  0.8383396349722769      1

    """
    features = features.copy()
    features_array = np.array(
        [
            features[col].data
            for col in features.columns
            if col not in ["obs_id", "dataset_name"]
        ]
    ).T

    default_linkage_kwargs = dict(method="ward", metric="euclidean")
    if linkage_kwargs is not None:
        default_linkage_kwargs.update(linkage_kwargs)

    pairwise_distances = sch.distance.pdist(features_array)
    linkage = sch.linkage(pairwise_distances, **default_linkage_kwargs)

    default_fcluster_kwargs = dict(criterion="maxclust", t=3)
    if fcluster_kwargs is not None:
        default_fcluster_kwargs.update(fcluster_kwargs)
    labels = sch.fcluster(linkage, **default_fcluster_kwargs)

    features["labels"] = labels
    return features



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import re
import numpy as np
from astropy.utils import lazyproperty
from gammapy.utils.array import is_power2
from ..utils import INVALID_INDEX

# Approximation of the size of HEALPix pixels (in degrees) for a particular order.
# Used to convert from HEALPix to WCS-based projections.
HPX_ORDER_TO_PIXSIZE = np.array(
    [32.0, 16.0, 8.0, 4.0, 2.0, 1.0, 0.50, 0.25, 0.1, 0.05, 0.025, 0.01, 0.005, 0.002]
)


def unravel_hpx_index(idx, npix):
    """Convert flattened global map index to an index tuple.

    Parameters
    ----------
    idx : `~numpy.ndarray`
        Flat index.
    npix : `~numpy.ndarray`
        Number of pixels in each band.

    Returns
    -------
    idx : tuple of `~numpy.ndarray`
        Index array for each dimension of the map.
    """
    if npix.size == 1:
        return tuple([idx])

    dpix = np.zeros(npix.size, dtype="i")
    dpix[1:] = np.cumsum(npix.flat[:-1])
    bidx = np.searchsorted(np.cumsum(npix.flat), idx + 1)
    pix = idx - dpix[bidx]
    return tuple([pix] + list(np.unravel_index(bidx, npix.shape)))


def ravel_hpx_index(idx, npix):
    """Convert map index tuple to a flattened index.

    Parameters
    ----------
    idx : tuple of `~numpy.ndarray`
        Index array.
    npix : `~numpy.ndarray`
        Number of pixels.

    Returns
    -------
    idx : `~numpy.ndarray`
        Index array.
    """
    if len(idx) == 1:
        return idx[0]

    # TODO: raise exception for indices that are out of bounds

    idx0 = idx[0]
    idx1 = np.ravel_multi_index(idx[1:], npix.shape, mode="clip")
    npix = np.concatenate((np.array([0]), npix.flat[:-1]))

    return idx0 + np.cumsum(npix)[idx1]


def coords_to_vec(lon, lat):
    """Convert longitude and latitude coordinates to a unit 3-vector.

    Returns
    -------
    array(3,n) with v_x[i],v_y[i],v_z[i] = directional cosines
    """
    phi = np.radians(lon)
    theta = (np.pi / 2) - np.radians(lat)
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)

    x = sin_t * np.cos(phi)
    y = sin_t * np.sin(phi)
    z = cos_t

    # Stack them into the output array
    out = np.vstack((x, y, z)).swapaxes(0, 1)
    return out


def get_nside_from_pix_size(pixsz):
    """Get the NSIDE that is closest to the given pixel size.

    Parameters
    ----------
    pixsz : `~numpy.ndarray`
        Pixel size in degrees.

    Returns
    -------
    nside : `~numpy.ndarray`
        HEALPix NSIDE parameter.
    """
    import healpy as hp

    pixsz = np.array(pixsz, ndmin=1)
    nside = 2 ** np.linspace(1, 14, 14, dtype=int)
    nside_pixsz = np.degrees(hp.nside2resol(nside))
    return nside[np.argmin(np.abs(nside_pixsz - pixsz[..., None]), axis=-1)]


def get_pix_size_from_nside(nside):
    """Estimate of the pixel size from the HEALPix nside coordinate.

    This just uses a lookup table to provide a nice round number
    for each HEALPix order.
    """
    order = nside_to_order(nside)
    if np.any(order < 0) or np.any(order > 13):
        raise ValueError(f"HEALPix order must be 0 to 13. Got: {order!r}")

    return HPX_ORDER_TO_PIXSIZE[order]


def match_hpx_pix(nside, nest, nside_pix, ipix_ring):
    """
    Match to the HEALPix pixel number.

    Parameters
    ----------
    nside : int or `~numpy.ndarray`
        HEALPix NSIDE parameter, must be a power of 2.
    nest : bool
        Indexing scheme. If True, "NESTED" scheme. If False, "RING" scheme.
    nside_pix : int or `~numpy.ndarray`
        HEALPix NSIDE parameter of subpixel.
    ipix_ring : int or `~numpy.ndarray`
        HEALPix pixel number.
    """
    import healpy as hp

    ipix_in = np.arange(12 * nside * nside)
    vecs = hp.pix2vec(nside, ipix_in, nest)
    pix_match = hp.vec2pix(nside_pix, vecs[0], vecs[1], vecs[2]) == ipix_ring
    return ipix_in[pix_match]


def parse_hpxregion(region):
    """Parse the ``HPX_REG`` header keyword into a list of tokens."""
    m = re.match(r"([A-Za-z\_]*?)\((.*?)\)", region)

    if m is None:
        raise ValueError(f"Failed to parse hpx region string: {region!r}")

    if not m.group(1):
        return re.split(",", m.group(2))
    else:
        return [m.group(1)] + re.split(",", m.group(2))


def nside_to_order(nside):
    """Compute the ORDER given NSIDE.

    Returns -1 for NSIDE values that are not a power of 2.
    """
    nside = np.array(nside, ndmin=1)
    order = -1 * np.ones_like(nside)
    m = is_power2(nside)
    order[m] = np.log2(nside[m]).astype(int)
    return order


def get_superpixels(idx, nside_subpix, nside_superpix, nest=True):
    """Compute the indices of superpixels that contain a subpixel.

    Parameters
    ----------
    idx : `~numpy.ndarray`
        Array of HEALPix pixel indices for subpixels of NSIDE
        ``nside_subpix``.
    nside_subpix : int or `~numpy.ndarray`
        HEALPix NSIDE parameter of subpixel.
    nside_superpix : int or `~numpy.ndarray`
        HEALPix NSIDE parameter of superpixel.
    nest : bool, optional
        Indexing scheme. If True, "NESTED" scheme. If False, "RING" scheme.
        Default is True.

    Returns
    -------
    idx_super : `~numpy.ndarray`
        Indices of HEALPix pixels of nside ``nside_superpix`` that
        contain pixel indices ``idx`` of nside ``nside_subpix``.
    """
    import healpy as hp

    idx = np.array(idx)
    nside_superpix = np.asarray(nside_superpix)
    nside_subpix = np.asarray(nside_subpix)

    if not nest:
        idx = hp.ring2nest(nside_subpix, idx)

    if np.any(~is_power2(nside_superpix)) or np.any(~is_power2(nside_subpix)):
        raise ValueError("NSIDE must be a power of 2.")

    ratio = np.array((nside_subpix // nside_superpix) ** 2, ndmin=1)
    idx //= ratio

    if not nest:
        m = idx == INVALID_INDEX.int
        idx[m] = 0
        idx = hp.nest2ring(nside_superpix, idx)
        idx[m] = INVALID_INDEX.int

    return idx


def get_subpixels(idx, nside_superpix, nside_subpix, nest=True):
    """Compute the indices of subpixels contained within superpixels.

    This function returns an output array with one additional
    dimension of size N for subpixel indices where N is the maximum
    number of subpixels for any pair of ``nside_superpix`` and
    ``nside_subpix``.  If the number of subpixels is less than N the
    remaining subpixel indices will be set to -1.

    Parameters
    ----------
    idx : `~numpy.ndarray`
        Array of HEALPix pixel indices for superpixels of NSIDE
        ``nside_superpix``.
    nside_superpix : int or `~numpy.ndarray`
        HEALPix NSIDE parameter of superpixel.
    nside_subpix : int or `~numpy.ndarray`
        HEALPix NSIDE parameter of subpixel.
    nest : bool, optional
        Indexing scheme. If True, "NESTED" scheme. If False, "RING" scheme.
        Default is True.

    Returns
    -------
    idx_sub : `~numpy.ndarray`
        Indices of HEALPix pixels of nside ``nside_subpix`` contained
        within pixel indices ``idx`` of nside ``nside_superpix``.
    """
    import healpy as hp

    if not nest:
        idx = hp.ring2nest(nside_superpix, idx)

    idx = np.asarray(idx)
    nside_superpix = np.asarray(nside_superpix)
    nside_subpix = np.asarray(nside_subpix)

    if np.any(~is_power2(nside_superpix)) or np.any(~is_power2(nside_subpix)):
        raise ValueError("NSIDE must be a power of 2.")

    # number of subpixels in each superpixel
    npix = np.array((nside_subpix // nside_superpix) ** 2, ndmin=1)
    x = np.arange(np.max(npix), dtype=int)
    idx = idx * npix

    if not np.all(npix[0] == npix):
        x = np.broadcast_to(x, idx.shape + x.shape)
        idx = idx[..., None] + x
        idx[x >= np.broadcast_to(npix[..., None], x.shape)] = INVALID_INDEX.int
    else:
        idx = idx[..., None] + x

    if not nest:
        m = idx == INVALID_INDEX.int
        idx[m] = 0
        idx = hp.nest2ring(nside_subpix[..., None], idx)
        idx[m] = INVALID_INDEX.int

    return idx


class HpxToWcsMapping:
    """Stores the indices need to convert from HEALPix to WCS.

    Parameters
    ----------
    hpx : `~HpxGeom`
        HEALPix geometry object.
    wcs : `~gammapy.maps.WcsGeom`
        WCS geometry object.
    """

    def __init__(self, hpx, wcs, ipix, mult_val, npix):
        self._hpx = hpx
        self._wcs = wcs
        self._ipix = ipix
        self._mult_val = mult_val
        self._npix = npix

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    @property
    def hpx(self):
        """HEALPix projection."""
        return self._hpx

    @property
    def wcs(self):
        """WCS projection."""
        return self._wcs

    @property
    def ipix(self):
        """An array(nx,ny) of the global HEALPix pixel indices for each WCS pixel."""
        return self._ipix

    @property
    def mult_val(self):
        """An array(nx,ny) of 1/number of WCS pixels pointing at each HEALPix pixel."""
        return self._mult_val

    @property
    def npix(self):
        """A tuple(nx,ny) of the shape of the WCS grid."""
        return self._npix

    @lazyproperty
    def lmap(self):
        """Array ``(nx, ny)`` mapping local HEALPix pixel indices for each WCS pixel."""
        return self.hpx.global_to_local(self.ipix, ravel=True)

    @property
    def valid(self):
        """Array ``(nx, ny)`` of bool: which WCS pixel in inside the HEALPix region."""
        return self.lmap >= 0

    @classmethod
    def create(cls, hpx, wcs):
        """Create HEALPix to WCS geometry pixel mapping.

        Parameters
        ----------
        hpx : `~HpxGeom`
            HEALPix geometry object.
        wcs : `~gammapy.maps.WcsGeom`
            WCS geometry object.

        Returns
        -------
        hpx2wcs : `~HpxToWcsMapping`
            Mapping.

        """
        import healpy as hp

        npix = wcs.npix

        # FIXME: Calculation of WCS pixel centers should be moved into a
        # method of WcsGeom
        pix_crds = np.dstack(np.meshgrid(np.arange(npix[0][0]), np.arange(npix[1][0])))
        pix_crds = pix_crds.swapaxes(0, 1).reshape((-1, 2))
        sky_crds = wcs.wcs.wcs_pix2world(pix_crds, 0)
        sky_crds *= np.radians(1.0)
        sky_crds[0:, 1] = (np.pi / 2) - sky_crds[0:, 1]

        mask = ~np.any(np.isnan(sky_crds), axis=1)
        ipix = -1 * np.ones((len(hpx.nside), int(npix[0][0] * npix[1][0])), int)
        m = mask[None, :] * np.ones_like(ipix, dtype=bool)

        ipix[m] = hp.ang2pix(
            hpx.nside[..., None],
            sky_crds[:, 1][mask][None, ...],
            sky_crds[:, 0][mask][None, ...],
            hpx.nest,
        ).flatten()

        # Here we are counting the number of HEALPix pixels each WCS pixel
        # points to and getting a multiplicative factor that tells use how
        # to split up the counts in each HEALPix pixel (by dividing the
        # corresponding WCS pixels by the number of associated HEALPix
        # pixels).
        mult_val = np.ones_like(ipix, dtype=float)
        for i, t in enumerate(ipix):
            count = np.unique(t, return_counts=True)
            idx = np.searchsorted(count[0], t)
            mult_val[i, ...] = 1.0 / count[1][idx]

        if hpx.nside.size == 1:
            ipix = np.squeeze(ipix, axis=0)
            mult_val = np.squeeze(mult_val, axis=0)

        return cls(hpx, wcs, ipix, mult_val, npix)

    def fill_wcs_map_from_hpx_data(
        self, hpx_data, wcs_data, normalize=True, fill_nan=True
    ):
        """Fill the WCS map from the HEALPix data using the pre-calculated mappings.

        Parameters
        ----------
        hpx_data : `~numpy.ndarray`
            The input HEALPix data.
        wcs_data : `~numpy.ndarray`
            The data array being filled.
        normalize : bool, optional
            If True, preserve integral by splitting HEALPix values between bins.
            Default is True.
        fill_nan : bool, optional
            Fill pixels outside the HEALPix geometry with NaN.
            Default is True.

        Returns
        -------
        wcs_data : `~numpy.ndarray`
            The WCS data array.
        """
        # FIXME: Do we want to flatten mapping arrays?

        shape = tuple([t.flat[0] for t in self._npix])
        if self.valid.ndim != 1:
            shape = hpx_data.shape[:-1] + shape

        valid = np.where(self.valid.reshape(shape))
        lmap = self.lmap[self.valid]
        mult_val = self._mult_val[self.valid]

        wcs_slice = [slice(None) for _ in range(wcs_data.ndim - 2)]
        wcs_slice = tuple(wcs_slice + list(valid)[::-1][:2])

        hpx_slice = [slice(None) for _ in range(wcs_data.ndim - 2)]
        hpx_slice = tuple(hpx_slice + [lmap])

        if normalize:
            wcs_data[wcs_slice] = mult_val * hpx_data[hpx_slice]
        else:
            wcs_data[wcs_slice] = hpx_data[hpx_slice]

        if fill_nan:
            valid = np.swapaxes(self.valid.reshape(shape), -1, -2)
            valid = valid * np.ones_like(wcs_data, dtype=bool)
            wcs_data[~valid] = np.nan

        return wcs_data

    def fill_hpx_map_from_wcs_data(self, wcs_data, hpx_data, normalize=True):
        """Fill the HEALPix map from the WCS data using the pre-calculated mappings.

        Parameters
        ----------
        wcs_data : `~numpy.ndarray`
            The input WCS data.
        hpx_data : `~numpy.ndarray`
            The data array being filled.
        normalize : bool, optional
            If True, preserve integral by splitting HEALPix values between bins.
            Default is True.

        Returns
        -------
        hpx_data : `~numpy.ndarray`
            The HEALPix data array.
        """
        shape = tuple([t.flat[0] for t in self._npix])
        if self.valid.ndim != 1:
            shape = hpx_data.shape[:-1] + shape

        valid = np.where(self.valid.reshape(shape))
        lmap = self.lmap[self.valid]
        mult_val = self._mult_val[self.valid]

        wcs_slice = [slice(None) for _ in range(wcs_data.ndim - 2)]
        wcs_slice = tuple(wcs_slice + list(valid)[::-1][:2])

        hpx_slice = [slice(None) for _ in range(wcs_data.ndim - 2)]
        hpx_slice = tuple(hpx_slice + [lmap])

        if normalize:
            hpx_data[hpx_slice] = 1 / mult_val * wcs_data[wcs_slice]
        else:
            hpx_data[hpx_slice] = wcs_data[wcs_slice]

        return hpx_data



"""
Makers - Data reduction
=======================

Data reduction: from observations to binned datasets

Introduction
------------

The `gammapy.makers` sub-package contains classes to perform data
reduction tasks from DL3 data to binned datasets. In the data reduction
step the DL3 data is prepared for modeling and fitting, by binning
events into a counts map and interpolating the exposure, background, psf
and energy dispersion on the chosen analysis geometry.

Setup
-----

"""

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion
import matplotlib.pyplot as plt
from gammapy.data import DataStore
from gammapy.datasets import Datasets, MapDataset, SpectrumDataset
from gammapy.makers import (
    DatasetsMaker,
    FoVBackgroundMaker,
    MapDatasetMaker,
    ReflectedRegionsBackgroundMaker,
    SafeMaskMaker,
    SpectrumDatasetMaker,
)
from gammapy.makers.utils import make_effective_livetime_map, make_observation_time_map
from gammapy.maps import MapAxis, RegionGeom, WcsGeom


######################################################################
# Dataset
# -------
#
# The counts, exposure, background and IRF maps are bundled together in a
# data structure named `~gammapy.datasets.MapDataset`.
#
# The first step of the data reduction is to create an empty dataset. A
# `~gammapy.datasets.MapDataset` can be created from any `~gammapy.maps.WcsGeom`
# object. This is illustrated in the following example:
#

energy_axis = MapAxis.from_bounds(
    1, 10, nbin=11, name="energy", unit="TeV", interp="log"
)
geom = WcsGeom.create(
    skydir=(83.63, 22.01),
    axes=[energy_axis],
    width=5 * u.deg,
    binsz=0.05 * u.deg,
    frame="icrs",
)
dataset_empty = MapDataset.create(geom=geom)
print(dataset_empty)


######################################################################
# It is possible to compute the instrument response functions with
# different spatial and energy bins as compared to the counts and
# background maps. For example, one can specify a true energy axis which
# defines the energy binning of the IRFs:
#

energy_axis_true = MapAxis.from_bounds(
    0.3, 10, nbin=31, name="energy_true", unit="TeV", interp="log"
)
dataset_empty = MapDataset.create(geom=geom, energy_axis_true=energy_axis_true)


######################################################################
# For the detail of the other options available, you can always call the
# help:
#

help(MapDataset.create)


######################################################################
# Once this empty “reference” dataset is defined, it can be filled with
# observational data using the `~gammapy.makers.MapDatasetMaker`:
#

# get observation
data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
obs = data_store.get_observations([23592])[0]

# fill dataset
maker = MapDatasetMaker()
dataset = maker.run(dataset_empty, obs)
print(dataset)

dataset.counts.sum_over_axes().plot(stretch="sqrt", add_cbar=True)
plt.show()


######################################################################
# The `~gammapy.makers.MapDatasetMaker` fills the corresponding ``counts``,
# ``exposure``, ``background``, ``psf`` and ``edisp`` map per observation.
# The `~gammapy.makers.MapDatasetMaker` has a ``selection`` parameter, in case some of
# the maps should not be computed. There is also a
# ``background_oversampling`` parameter that defines the oversampling
# factor in energy used to compute the background (default is None).
#
# .. _safe-data-range:
#
# Safe data range handling
# ------------------------
#
# To exclude the data range from a `~gammapy.datasets.MapDataset`, that is associated with
# high systematics on instrument response functions, a `~gammapy.datasets.MapDataset.mask_safe`
# can be defined. The `~gammapy.datasets.MapDataset.mask_safe` is a `~gammapy.maps.Map` object
# with `bool` data type, which indicates for each pixel, whether it should be included in
# the analysis. The convention is that a value of `True` or `1`
# includes the pixel, while a value of `False` or `0` excludes a
# pixels from the analysis. To compute safe data range masks according to
# certain criteria, Gammapy provides a `~gammapy.makers.SafeMaskMaker` class. The
# different criteria are given by the `methods` argument, available
# options are :
#
# -  aeff-default, uses the energy ranged specified in the DL3 data files,
#    if available.
# -  aeff-max, the lower energy threshold is determined such as the
#    effective area is above a given percentage of its maximum
# -  edisp-bias, the lower energy threshold is determined such as the
#    energy bias is below a given percentage
# -  offset-max, the data beyond a given offset radius from the
#    observation center are excluded
# -  bkg-peak, the energy threshold is defined as the upper edge of the
#    energy bin with the highest predicted background rate. This method
#    was introduced in the
#    `H.E.S.S. DL3 validation paper <https://arxiv.org/pdf/1910.08088.pdf>`__
#
# Note that currently some methods computing a safe energy range
# ("aeff-default", "aeff-max" and "edisp-bias") determine a true energy range and
# apply it to reconstructed energy, effectively neglecting the energy dispersion.
#
# Multiple methods can be combined. Here is an example :
#

safe_mask_maker = SafeMaskMaker(
    methods=["aeff-default", "offset-max"], offset_max="3 deg"
)

dataset = maker.run(dataset_empty, obs)
dataset = safe_mask_maker.run(dataset, obs)
print(dataset.mask_safe)

dataset.mask_safe.sum_over_axes().plot()
plt.show()


######################################################################
# The `~gammapy.makers.SafeMaskMaker` does not modify any data, but only defines the
# `~gammapy.datasets.MapDataset.mask_safe` attribute. This means that the safe data range
# can be defined and modified in between the data reduction and stacking
# and fitting. For a joint-likelihood analysis of multiple observations
# the safe mask is applied to the counts and predicted number of counts
# map during fitting. This correctly accounts for contributions
# (spill-over) by the PSF from outside the field of view.
#
# Background estimation
# ---------------------
#
# The background computed by the `~gammapy.makers.MapDatasetMaker` gives the number of
# counts predicted by the background IRF of the observation. Because its
# actual normalization, or even its spectral shape, might be poorly
# constrained, it is necessary to correct it with the data themselves.
# This is the role of background estimation Makers.
#
# FoV background
# ~~~~~~~~~~~~~~
#
# If the background energy dependent morphology is not well reproduced by the
# background model stored in the IRF, it might be that its normalization
# is incorrect and that some spectral corrections are necessary. This is
# made possible thanks to the `~gammapy.makers.FoVBackgroundMaker`. This
# technique is recommended in most 3D data reductions. For more details
# and usage, see the :doc:`FoV background </user-guide/makers/fov>`.
#
# Here we are going to use a `~gammapy.makers.FoVBackgroundMaker` that
# will rescale the background model to the data excluding the region where
# a known source is present. For more details on the way to create
# exclusion masks see the :doc:`mask maps </tutorials/details/mask_maps>` notebook.
#

circle = CircleSkyRegion(center=geom.center_skydir, radius=0.2 * u.deg)
exclusion_mask = geom.region_mask([circle], inside=False)

fov_bkg_maker = FoVBackgroundMaker(method="scale", exclusion_mask=exclusion_mask)
dataset = fov_bkg_maker.run(dataset)


######################################################################
# Other backgrounds production methods available are listed below.
#
# Ring background
# ~~~~~~~~~~~~~~~
#
# If the background model does not reproduce well the morphology, a
# classical approach consists in applying local corrections by smoothing
# the data with a ring kernel. This allows to build a set of OFF counts
# taking into account the imperfect knowledge of the background. This is
# implemented in the `~gammapy.makers.RingBackgroundMaker` which
# transforms the Dataset in a `~gammapy.datasets.MapDatasetOnOff`. This technique is
# mostly used for imaging, and should not be applied for 3D modeling and
# fitting.
#
# For more details and usage, see
# :doc:`Ring background </user-guide/makers/ring>`
#
# Reflected regions background
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the absence of a solid background model, a classical technique in
# Cherenkov astronomy for 1D spectral analysis is to estimate the
# background in a number of OFF regions. When the background can be safely
# estimated as radially symmetric w.r.t. the pointing direction, one can
# apply the reflected regions background technique. This is implemented in
# the `~gammapy.makers.ReflectedRegionsBackgroundMaker` which transforms
# a `~gammapy.datasets.SpectrumDataset` in a `~gammapy.datasets.SpectrumDatasetOnOff`.
# This method is only used for 1D spectral analysis.
#
# For more details and usage, see
# the :doc:`Reflected background </user-guide/makers/reflected>`
#
# Data reduction loop
# -------------------
#
# The data reduction steps can be combined in a single loop to run a full
# data reduction chain. For this the `~gammapy.makers.MapDatasetMaker` is run first and
# the output dataset is the passed on to the next maker step. Finally, the
# dataset per observation is stacked into a larger map.
#

data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
observations = data_store.get_observations([23523, 23592, 23526, 23559])

energy_axis = MapAxis.from_bounds(
    1, 10, nbin=11, name="energy", unit="TeV", interp="log"
)
geom = WcsGeom.create(skydir=(83.63, 22.01), axes=[energy_axis], width=5, binsz=0.02)

dataset_maker = MapDatasetMaker()
safe_mask_maker = SafeMaskMaker(
    methods=["aeff-default", "offset-max"], offset_max="3 deg"
)

stacked = MapDataset.create(geom)

for obs in observations:
    local_dataset = stacked.cutout(obs.get_pointing_icrs(obs.tmid), width="6 deg")
    dataset = dataset_maker.run(local_dataset, obs)
    dataset = safe_mask_maker.run(dataset, obs)
    dataset = fov_bkg_maker.run(dataset)
    stacked.stack(dataset)

print(stacked)


######################################################################
# To maintain good performance it is always recommended to do a cutout of
# the `~gammapy.datasets.MapDataset` as shown above. In case you want to increase the
# offset-cut later, you can also choose a larger width of the cutout than
# ``2 * offset_max``.
#
# Note that we stack the individual `~gammapy.datasets.MapDataset`, which are computed per
# observation into a larger dataset. During the stacking the safe data
# range mask (`~gammapy.datasets.MapDataset.mask_safe`) is applied by setting data outside
# to zero, then data is added to the larger map dataset. To stack multiple
# observations, the larger dataset must be created first.
#
# The data reduction loop shown above can be done through the
# `~gammapy.makers.DatasetsMaker` class that take as argument a list of makers. **Note
# that the order of the makers list is important as it determines their
# execution order.** Moreover the ``stack_datasets`` option offers the
# possibility to stack or not the output datasets, and the ``n_jobs`` option
# allow to use multiple processes on run.
#

global_dataset = MapDataset.create(geom)
makers = [dataset_maker, safe_mask_maker, fov_bkg_maker]  # the order matter
datasets_maker = DatasetsMaker(makers, stack_datasets=False, n_jobs=1)
datasets = datasets_maker.run(global_dataset, observations)
print(datasets)


######################################################################
# Spectrum dataset
# ----------------
#
# The spectrum datasets represent 1D spectra along an energy axis within a
# given on region. The `~gammapy.datasets.SpectrumDataset` contains a counts spectrum, and
# a background model. The `~gammapy.datasets.SpectrumDatasetOnOff` contains ON and OFF
# count spectra, background is implicitly modeled via the OFF counts
# spectrum.
#
# The `~gammapy.makers.SpectrumDatasetMaker` make spectrum dataset for a single
# observation. In that case the IRFs and background are computed at a
# single fixed offset, which is recommended only for point-sources.
#
# Here is an example of data reduction loop to create
# `~gammapy.datasets.SpectrumDatasetOnOff` datasets:
#

# on region is given by the CircleSkyRegion previously defined
geom = RegionGeom.create(region=circle, axes=[energy_axis])
exclusion_mask_2d = exclusion_mask.reduce_over_axes(np.logical_or, keepdims=False)

spectrum_dataset_empty = SpectrumDataset.create(
    geom=geom, energy_axis_true=energy_axis_true
)

spectrum_dataset_maker = SpectrumDatasetMaker(
    containment_correction=False, selection=["counts", "exposure", "edisp"]
)
reflected_bkg_maker = ReflectedRegionsBackgroundMaker(exclusion_mask=exclusion_mask_2d)
safe_mask_masker = SafeMaskMaker(methods=["aeff-max"], aeff_percent=10)

datasets = Datasets()

for observation in observations:
    dataset = spectrum_dataset_maker.run(
        spectrum_dataset_empty.copy(name=f"obs-{observation.obs_id}"),
        observation,
    )
    dataset_on_off = reflected_bkg_maker.run(dataset, observation)
    dataset_on_off = safe_mask_masker.run(dataset_on_off, observation)
    datasets.append(dataset_on_off)
print(datasets)

######################################################################
# Observation duration and effective livetime
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# It can often be useful to know the total number of hours spent
# in the given field of view (without correcting for the acceptance
# variation). This can be computed using `~gammapy.makers.utils.make_observation_time_map`
# as shown below
#

# Get the observations
obs_id = data_store.obs_table["OBS_ID"][data_store.obs_table["OBJECT"] == "MSH 15-5-02"]
observations = data_store.get_observations(obs_id)
print("No. of observations: ", len(observations))

# Define an energy range
energy_min = 100 * u.GeV
energy_max = 10.0 * u.TeV

# Define an offset cut (the camera field of view)
offset_max = 2.5 * u.deg

# Define the geom
source_pos = SkyCoord(228.32, -59.08, unit="deg")
energy_axis_true = MapAxis.from_energy_bounds(
    energy_min, energy_max, nbin=2, name="energy_true"
)
geom = WcsGeom.create(
    skydir=source_pos,
    binsz=0.02,
    width=(6, 6),
    frame="icrs",
    proj="CAR",
    axes=[energy_axis_true],
)

total_obstime = make_observation_time_map(observations, geom, offset_max=offset_max)


plt.figure(figsize=(5, 5))
ax = total_obstime.plot(add_cbar=True)
# Add the pointing position on top
for obs in observations:
    ax.plot(
        obs.get_pointing_icrs(obs.tmid).to_pixel(wcs=ax.wcs)[0],
        obs.get_pointing_icrs(obs.tmid).to_pixel(wcs=ax.wcs)[1],
        "+",
        color="black",
    )
ax.set_title("Total observation time")
plt.show()

######################################################################
# As the acceptance of IACT cameras vary within the field of
# view, it can also be interesting to plot the on-axis equivalent
# number of hours.
#

effective_livetime = make_effective_livetime_map(
    observations, geom, offset_max=offset_max
)


axs = effective_livetime.plot_grid(add_cbar=True)
# Add the pointing position on top
for ax in axs:
    for obs in observations:
        ax.plot(
            obs.get_pointing_icrs(obs.tmid).to_pixel(wcs=ax.wcs)[0],
            obs.get_pointing_icrs(obs.tmid).to_pixel(wcs=ax.wcs)[1],
            "+",
            color="black",
        )
plt.show()

######################################################################
# To get the value of the observation time at a particular position,
# use ``get_by_coord``

obs_time_src = total_obstime.get_by_coord(source_pos)
effective_times_src = effective_livetime.get_by_coord(
    (source_pos, energy_axis_true.center)
)

print(f"Time spent on position {source_pos}")
print(f"Total observation time: {obs_time_src}* {total_obstime.unit}")
print(
    f"Effective livetime at {energy_axis_true.center[0]}: {effective_times_src[0]} * {effective_livetime.unit}"
)
print(
    f"Effective livetime at {energy_axis_true.center[1]}: {effective_times_src[1]} * {effective_livetime.unit}"
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from copy import deepcopy
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion
from gammapy.datasets.evaluator import MapEvaluator
from gammapy.irf import PSFKernel, RecoPSFMap
from gammapy.maps import Map, MapAxis, RegionGeom, RegionNDMap, WcsGeom
from gammapy.modeling.models import (
    ConstantSpectralModel,
    GaussianSpatialModel,
    Models,
    PointSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)
from gammapy.utils.gauss import Gauss2DPDF
from gammapy.utils.testing import mpl_plot_check


@pytest.fixture
def evaluator():
    center = SkyCoord("0 deg", "0 deg", frame="galactic")
    region = CircleSkyRegion(center=center, radius=0.1 * u.deg)

    nbin = 2
    energy_axis_true = MapAxis.from_energy_bounds(
        ".1 TeV", "10 TeV", nbin=nbin, name="energy_true"
    )

    spectral_model = ConstantSpectralModel()
    spatial_model = PointSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, frame="galactic"
    )

    models = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)
    model = Models(models)

    exposure_region = RegionNDMap.create(
        region, axes=[energy_axis_true], binsz_wcs="0.01deg", unit="m2 s", data=1.0
    )

    geom = RegionGeom(region, axes=[energy_axis_true], binsz_wcs="0.01deg")
    psf = PSFKernel.from_gauss(geom.to_wcs_geom(), sigma="0.1 deg")

    return MapEvaluator(model=model[0], exposure=exposure_region, psf=psf)


def test_compute_flux_spatial(evaluator):
    flux = evaluator.compute_flux_spatial()

    g = Gauss2DPDF(0.1)
    reference = g.containment_fraction(0.1)
    assert_allclose(flux.value, reference, rtol=0.003)


def test_compute_flux_spatial_no_psf(evaluator):
    # check that spatial integration is not performed in the absence of a psf

    evaluator_gaussian = deepcopy(evaluator)
    model = evaluator.model.copy()
    model.spatial_model = GaussianSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, sigma=0.001 * u.deg, frame="galactic"
    )
    evaluator_gaussian.model = model

    evaluator_gaussian.model.apply_irf["psf"] = False
    flux = evaluator_gaussian.compute_flux_spatial()
    evaluator_gaussian.model.apply_irf["psf"] = True
    assert_allclose(flux, 1.0)

    evaluator_gaussian.psf = None
    flux = evaluator_gaussian.compute_flux_spatial()
    assert_allclose(flux, 1.0)

    evaluator.model.apply_irf["psf"] = False
    flux = evaluator.compute_flux_spatial()
    evaluator.model.apply_irf["psf"] = True
    assert_allclose(flux, 1.0)

    evaluator.psf = None
    flux = evaluator.compute_flux_spatial()
    assert_allclose(flux, 1.0)


def test_large_oversampling():
    nbin = 2
    energy_axis_true = MapAxis.from_energy_bounds(
        ".1 TeV", "10 TeV", nbin=nbin, name="energy_true"
    )
    geom = WcsGeom.create(width=1, binsz=0.02, axes=[energy_axis_true])

    spectral_model = ConstantSpectralModel()
    spatial_model = GaussianSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, sigma=1e-4 * u.deg, frame="icrs"
    )

    models = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)
    model = Models(models)

    exposure = Map.from_geom(geom, unit="m2 s")
    exposure.data += 1.0

    psf = PSFKernel.from_gauss(geom, sigma="0.1 deg")

    evaluator = MapEvaluator(model=model[0], exposure=exposure, psf=psf)
    flux_1 = evaluator.compute_flux_spatial()

    spatial_model.sigma.value = 0.001
    flux_2 = evaluator.compute_flux_spatial()

    spatial_model.sigma.value = 0.01
    flux_3 = evaluator.compute_flux_spatial()

    spatial_model.sigma.value = 0.03
    flux_4 = evaluator.compute_flux_spatial()

    assert_allclose(flux_1.data.sum(), nbin, rtol=1e-4)
    assert_allclose(flux_2.data.sum(), nbin, rtol=1e-4)
    assert_allclose(flux_3.data.sum(), nbin, rtol=1e-4)
    assert_allclose(flux_4.data.sum(), nbin, rtol=1e-4)


def test_compute_npred_sign():
    center = SkyCoord("0 deg", "0 deg", frame="galactic")
    energy_axis_true = MapAxis.from_energy_bounds(
        ".1 TeV", "10 TeV", nbin=2, name="energy_true"
    )
    geom = WcsGeom.create(
        skydir=center,
        width=1 * u.deg,
        axes=[energy_axis_true],
        frame="galactic",
        binsz=0.2 * u.deg,
    )

    spectral_model_pos = PowerLawSpectralModel(index=2, amplitude="1e-11 TeV-1 s-1 m-2")
    spectral_model_neg = PowerLawSpectralModel(
        index=2, amplitude="-1e-11 TeV-1 s-1 m-2"
    )

    spatial_model = PointSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, frame="galactic"
    )
    model_pos = SkyModel(spectral_model=spectral_model_pos, spatial_model=spatial_model)
    model_neg = SkyModel(spectral_model=spectral_model_neg, spatial_model=spatial_model)

    exposure = Map.from_geom(geom, unit="m2 s")
    exposure.data += 1.0

    psf = PSFKernel.from_gauss(geom, sigma="0.1 deg")

    evaluator_pos = MapEvaluator(model=model_pos, exposure=exposure, psf=psf)
    evaluator_neg = MapEvaluator(model=model_neg, exposure=exposure, psf=psf)

    npred_pos = evaluator_pos.compute_npred()
    npred_neg = evaluator_neg.compute_npred()

    assert (npred_pos.data == -npred_neg.data).all()
    assert np.all(npred_pos.data >= 0)
    assert np.all(npred_neg.data <= 0)


def test_psf_reco():
    center = SkyCoord("0 deg", "0 deg", frame="galactic")
    energy_axis = MapAxis.from_energy_bounds("1 TeV", "10 TeV", nbin=3, name="energy")
    geom = WcsGeom.create(
        skydir=center,
        width=1 * u.deg,
        axes=[energy_axis],
        frame="galactic",
        binsz=0.2 * u.deg,
    )

    spectral_model = PowerLawSpectralModel(index=2, amplitude="1e-11 TeV-1 s-1 m-2")

    spatial_model = PointSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, frame="galactic"
    )
    model_pos = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)

    exposure = Map.from_geom(geom.as_energy_true, unit="m2 s")
    exposure.data += 1.0

    psf = PSFKernel.from_gauss(geom, sigma=0.1 * u.deg)

    evaluator = MapEvaluator(model=model_pos, exposure=exposure, psf=psf)

    assert evaluator.apply_psf_after_edisp
    assert evaluator.methods_sequence[-1] == evaluator.apply_psf
    npred = evaluator.compute_npred()
    assert_allclose(npred.data.sum(), 9e-12)

    psf_map = RecoPSFMap.from_gauss(
        energy_axis=energy_axis, sigma=0.1 * u.deg, geom=geom.to_image()
    )

    mask = Map.from_geom(geom, data=True)
    evaluator.psf = None
    evaluator.update(exposure, psf_map, None, geom, mask)
    assert evaluator.apply_psf_after_edisp
    assert evaluator.methods_sequence[-1] == evaluator.apply_psf
    assert_allclose(evaluator.compute_npred().data.sum(), 9e-12)


def test_peek_region_geom(evaluator):
    with mpl_plot_check():
        evaluator.peek()


def test_peek_wcs_geom():
    center = SkyCoord("0 deg", "0 deg", frame="galactic")
    energy_axis_true = MapAxis.from_energy_bounds(
        ".1 TeV", "10 TeV", nbin=2, name="energy_true"
    )
    geom = WcsGeom.create(
        skydir=center,
        width=1 * u.deg,
        axes=[energy_axis_true],
        frame="galactic",
        binsz=0.2 * u.deg,
    )

    spectral_model = PowerLawSpectralModel()
    spatial_model = PointSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, frame="galactic"
    )
    model = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)

    exposure = Map.from_geom(geom, unit="m2 s")
    exposure.data += 1.0

    evaluator = MapEvaluator(model=model, exposure=exposure)

    with mpl_plot_check():
        evaluator.peek()


def test_norm_only_changed():
    center = SkyCoord("0 deg", "0 deg", frame="galactic")
    energy_axis_true = MapAxis.from_energy_bounds(
        ".1 TeV", "10 TeV", nbin=2, name="energy_true"
    )
    geom = WcsGeom.create(
        skydir=center,
        width=1 * u.deg,
        axes=[energy_axis_true],
        frame="galactic",
        binsz=0.2 * u.deg,
    )

    spectral_model = PowerLawSpectralModel(index=2, amplitude="1e-11 TeV-1 s-1 m-2")

    spatial_model = PointSpatialModel(
        lon_0=0 * u.deg, lat_0=0 * u.deg, frame="galactic"
    )
    model = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model)

    exposure = Map.from_geom(geom, unit="m2 s")
    exposure.data += 1.0

    psf = PSFKernel.from_gauss(geom, sigma="0.1 deg")

    evaluator = MapEvaluator(model=model, exposure=exposure, psf=psf)

    _ = evaluator.compute_npred()

    spectral_model.amplitude.value *= 2
    assert evaluator.parameter_norm_only_changed

    spectral_model.index.value *= 2
    assert not evaluator.parameter_norm_only_changed

    spectral_model.amplitude.value *= 2
    spectral_model.index.value *= 2
    assert not evaluator.parameter_norm_only_changed



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
from pathlib import Path
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion
from pydantic import ValidationError
from gammapy.analysis import Analysis, AnalysisConfig
from gammapy.datasets import MapDataset, SpectrumDatasetOnOff
from gammapy.maps import WcsGeom, WcsNDMap
from gammapy.modeling.models import DatasetModels
from gammapy.utils.testing import requires_data

CONFIG_PATH = Path(__file__).resolve().parent / ".." / "config"
MODEL_FILE = CONFIG_PATH / "model.yaml"
MODEL_FILE_1D = CONFIG_PATH / "model-1d.yaml"


def get_example_config(which):
    """Example config: which can be 1d or 3d."""
    return AnalysisConfig.read(CONFIG_PATH / f"example-{which}.yaml")


def test_init():
    cfg = {"general": {"outdir": "test"}}
    analysis = Analysis(cfg)
    assert analysis.config.general.outdir == "test"
    with pytest.raises(TypeError):
        Analysis("spam")


def test_update_config():
    analysis = Analysis(AnalysisConfig())
    data = {"general": {"outdir": "test"}}
    config = AnalysisConfig(**data)
    analysis.update_config(config)
    assert analysis.config.general.outdir == "test"

    analysis = Analysis(AnalysisConfig())
    data = """
    general:
        outdir: test
    """
    analysis.update_config(data)
    assert analysis.config.general.outdir == "test"

    analysis = Analysis(AnalysisConfig())
    with pytest.raises(TypeError):
        analysis.update_config(0)


def test_get_observations_no_datastore():
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.datastore = "other"
    with pytest.raises(FileNotFoundError):
        analysis.get_observations()


@requires_data()
def test_get_observations_all():
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.datastore = "$GAMMAPY_DATA/cta-1dc/index/gps/"
    analysis.get_observations()
    assert len(analysis.observations) == 4


@requires_data()
def test_get_observations_obs_ids():
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.datastore = "$GAMMAPY_DATA/cta-1dc/index/gps/"
    analysis.config.observations.obs_ids = ["110380"]
    analysis.get_observations()
    assert len(analysis.observations) == 1


@requires_data()
def test_get_observations_obs_cone():
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.datastore = "$GAMMAPY_DATA/hess-dl3-dr1"
    analysis.config.observations.obs_cone = {
        "frame": "icrs",
        "lon": "83d",
        "lat": "22d",
        "radius": "5d",
    }
    analysis.get_observations()
    assert len(analysis.observations) == 4


@requires_data()
def test_get_observations_obs_file(tmp_path):
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.get_observations()
    filename = tmp_path / "obs_ids.txt"
    filename.write_text("20136\n47829\n")
    analysis.config.observations.obs_file = filename
    analysis.get_observations()
    assert len(analysis.observations) == 2


@requires_data()
def test_get_observations_obs_time(tmp_path):
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.obs_time = {
        "start": "2004-03-26",
        "stop": "2004-05-26",
    }
    analysis.get_observations()
    assert len(analysis.observations) == 40
    analysis.config.observations.obs_ids = [0]
    with pytest.raises(KeyError):
        analysis.get_observations()


@requires_data()
def test_get_observations_missing_irf():
    config = AnalysisConfig()
    analysis = Analysis(config)
    analysis.config.observations.datastore = "$GAMMAPY_DATA/joint-crab/dl3/magic/"
    analysis.config.observations.obs_ids = ["05029748"]
    analysis.config.observations.required_irf = ["aeff", "edisp"]
    analysis.get_observations()
    assert len(analysis.observations) == 1


@requires_data()
def test_set_models():
    config = get_example_config("3d")
    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    models_str = Path(MODEL_FILE).read_text()
    analysis.set_models(models=models_str)
    assert isinstance(analysis.models, DatasetModels)
    assert len(analysis.models) == 2
    assert analysis.models.names == ["source", "stacked-bkg"]
    with pytest.raises(TypeError):
        analysis.set_models(0)

    new_source = analysis.models["source"].copy(name="source2")
    analysis.set_models(models=[new_source], extend=False)
    assert len(analysis.models) == 2
    assert analysis.models.names == ["source2", "stacked-bkg"]


@requires_data()
def test_analysis_1d():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523, 23526]
        obs_time: {
            start: [J2004.92654346, J2004.92658453, J2004.92663655],
            stop: [J2004.92658453, J2004.92663655, J2004.92670773]
        }
    datasets:
        type: 1d
        background:
            method: reflected
        geom:
            axes:
                energy_true: {min: 0.01 TeV, max: 300 TeV, nbins: 109}
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.11 deg}
        safe_mask:
            methods: [aeff-default, edisp-bias]
            parameters: {bias_percent: 10.0}
        containment_correction: false
    flux_points:
        energy: {min: 1 TeV, max: 50 TeV, nbins: 4}
    light_curve:
        energy_edges: {min: 1 TeV, max: 50 TeV, nbins: 1}
        time_intervals: {
            start: [J2004.92654346, J2004.92658453, J2004.92663655],
            stop: [J2004.92658453, J2004.92663655, J2004.92670773]
        }
    """
    config = get_example_config("1d")
    analysis = Analysis(config)
    analysis.update_config(cfg)
    analysis.get_observations()
    analysis.get_datasets()
    analysis.read_models(MODEL_FILE_1D)
    analysis.run_fit()
    analysis.get_flux_points()
    analysis.get_light_curve()

    assert len(analysis.datasets) == 3
    table = analysis.flux_points.data.to_table(sed_type="dnde")

    assert len(table) == 4
    dnde = table["dnde"].quantity
    assert dnde.unit == "cm-2 s-1 TeV-1"

    assert_allclose(dnde[0].value, 8.116854e-12, rtol=1e-2)
    assert_allclose(dnde[2].value, 3.444475e-14, rtol=1e-2)

    axis = analysis.light_curve.geom.axes["time"]
    assert axis.nbin == 3
    assert_allclose(axis.time_min.mjd, [53343.92, 53343.935, 53343.954])

    flux = analysis.light_curve.flux.data[:, :, 0, 0]
    assert_allclose(flux, [[1.688954e-11], [2.347870e-11], [1.604152e-11]], rtol=1e-4)


@requires_data()
def test_geom_analysis_1d():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_ids: [23523]
    datasets:
        type: 1d
        background:
            method: reflected
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.11 deg}
        geom:
            axes:
                energy: {min: 0.1 TeV, max: 30 TeV, nbins: 20}
                energy_true: {min: 0.03 TeV, max: 100 TeV, nbins: 50}
        containment_correction: false
    flux_points:
        energy: {min: 1 TeV, max: 50 TeV, nbins: 4}
    """
    config = get_example_config("1d")
    analysis = Analysis(config)
    analysis.update_config(cfg)
    analysis.get_observations()
    analysis.get_datasets()

    assert len(analysis.datasets) == 1

    axis = analysis.datasets[0].exposure.geom.axes["energy_true"]
    assert axis.nbin == 50
    assert_allclose(axis.edges[0].to_value("TeV"), 0.03)
    assert_allclose(axis.edges[-1].to_value("TeV"), 100)


@requires_data()
def test_exclusion_region(tmp_path):
    config = get_example_config("1d")
    analysis = Analysis(config)
    region = CircleSkyRegion(center=SkyCoord("85d 23d"), radius=1 * u.deg)
    geom = WcsGeom.create(npix=(150, 150), binsz=0.05, skydir=SkyCoord("83d 22d"))
    exclusion_mask = ~geom.region_mask([region])

    filename = tmp_path / "exclusion.fits"
    exclusion_mask.write(filename)
    config.datasets.background.method = "reflected"
    config.datasets.background.exclusion = filename
    analysis.get_observations()
    analysis.get_datasets()
    assert len(analysis.datasets) == 2

    config = get_example_config("3d")
    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    geom = analysis.datasets[0]._geom
    exclusion_mask = ~geom.region_mask([region])
    filename = tmp_path / "exclusion3d.fits"
    exclusion_mask.write(filename)
    config.datasets.background.exclusion = filename
    analysis.get_datasets()
    assert len(analysis.datasets) == 1


@requires_data()
def test_analysis_1d_stacked_no_fit_range():
    cfg = """
    observations:
        datastore: $GAMMAPY_DATA/hess-dl3-dr1
        obs_cone: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 5 deg}
        obs_ids: [23592, 23559]

    datasets:
        type: 1d
        stack: false
        geom:
            axes:
                energy: {min: 0.01 TeV, max: 100 TeV, nbins: 73}
                energy_true: {min: 0.03 TeV, max: 100 TeV, nbins: 50}
        on_region: {frame: icrs, lon: 83.633 deg, lat: 22.014 deg, radius: 0.1 deg}
        containment_correction: true
        background:
            method: reflected
    """
    config = AnalysisConfig.from_yaml(cfg)
    analysis = Analysis(config)
    analysis.update_config(cfg)
    analysis.config.datasets.stack = True
    analysis.get_observations()
    analysis.get_datasets()
    analysis.read_models(MODEL_FILE_1D)
    analysis.run_fit()
    with pytest.raises(ValueError):
        analysis.get_excess_map()

    assert len(analysis.datasets) == 1
    assert_allclose(analysis.datasets["stacked"].counts.data.sum(), 184)
    pars = analysis.models.parameters
    assert_allclose(analysis.datasets[0].mask_fit.data, True)

    assert_allclose(pars["index"].value, 2.76913, rtol=1e-2)
    assert_allclose(pars["amplitude"].value, 5.479729e-11, rtol=1e-2)


@requires_data()
def test_analysis_ring_background():
    config = get_example_config("3d")
    config.datasets.background.method = "ring"
    config.datasets.background.parameters = {"r_in": "0.7 deg", "width": "0.7 deg"}
    config.datasets.geom.axes.energy.nbins = 1
    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    analysis.get_excess_map()
    assert isinstance(analysis.datasets[0], MapDataset)
    assert_allclose(
        analysis.datasets[0].npred_background().data[0, 10, 10], 0.091799, rtol=1e-2
    )
    assert isinstance(analysis.excess_map["sqrt_ts"], WcsNDMap)
    assert_allclose(analysis.excess_map.npred_excess.data[0, 62, 62], 134.12389)


@requires_data()
def test_analysis_ring_3d():
    config = get_example_config("3d")
    config.datasets.background.method = "ring"
    config.datasets.background.parameters = {"r_in": "0.7 deg", "width": "0.7 deg"}
    analysis = Analysis(config)
    analysis.get_observations()
    with pytest.raises(ValueError):
        analysis.get_datasets()


@requires_data()
def test_analysis_no_bkg_1d(caplog):
    config = get_example_config("1d")
    analysis = Analysis(config)
    with caplog.at_level(logging.WARNING):
        analysis.get_observations()
        analysis.get_datasets()
        assert not isinstance(analysis.datasets[0], SpectrumDatasetOnOff)
        assert "No background maker set. Check configuration." in [
            _.message for _ in caplog.records
        ]


@requires_data()
def test_analysis_no_bkg_3d(caplog):
    config = get_example_config("3d")
    config.datasets.background.method = None
    analysis = Analysis(config)
    with caplog.at_level(logging.WARNING):
        analysis.get_observations()
        analysis.get_datasets()
        assert isinstance(analysis.datasets[0], MapDataset)
        assert "No background maker set. Check configuration." in [
            _.message for _ in caplog.records
        ]


@requires_data()
def test_analysis_3d():
    config = get_example_config("3d")
    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    analysis.read_models(MODEL_FILE)
    analysis.datasets["stacked"].background_model.spectral_model.tilt.frozen = False
    analysis.run_fit()
    analysis.get_flux_points()

    assert len(analysis.datasets) == 1
    assert len(analysis.models.parameters) == 8
    res = analysis.models.parameters
    assert res["amplitude"].unit == "cm-2 s-1 TeV-1"

    table = analysis.flux_points.data.to_table(sed_type="dnde")
    assert len(table) == 2
    dnde = table["dnde"].quantity

    assert_allclose(dnde[0].value, 1.2722e-11, rtol=1e-2)
    assert_allclose(dnde[-1].value, 4.054128e-13, rtol=1e-2)
    assert_allclose(res["index"].value, 2.772814, rtol=1e-2)
    assert_allclose(res["tilt"].value, -0.133436, rtol=1e-2)


@requires_data()
def test_analysis_3d_joint_datasets():
    config = get_example_config("3d")
    config.datasets.stack = False
    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    assert len(analysis.datasets) == 2

    assert_allclose(
        analysis.datasets[0].background_model.spectral_model.norm.value,
        1.031743694988066,
        rtol=1e-6,
    )
    assert_allclose(
        analysis.datasets[0].background_model.spectral_model.tilt.value,
        0.0,
        rtol=1e-6,
    )
    assert_allclose(
        analysis.datasets[1].background_model.spectral_model.norm.value,
        0.9776349021876344,
        rtol=1e-6,
    )


@requires_data()
def test_usage_errors():
    config = get_example_config("1d")
    analysis = Analysis(config)
    with pytest.raises(RuntimeError):
        analysis.get_datasets()
    with pytest.raises(RuntimeError):
        analysis.read_datasets()
    with pytest.raises(RuntimeError):
        analysis.write_datasets()
    with pytest.raises(TypeError):
        analysis.read_models()
    with pytest.raises(RuntimeError):
        analysis.write_models()
    with pytest.raises(RuntimeError):
        analysis.run_fit()
    with pytest.raises(RuntimeError):
        analysis.get_flux_points()
    with pytest.raises(ValidationError):
        analysis.config.datasets.type = "None"


@requires_data()
def test_datasets_io(tmpdir):
    config = get_example_config("3d")

    analysis = Analysis(config)
    analysis.get_observations()
    analysis.get_datasets()
    models_str = Path(MODEL_FILE).read_text()
    analysis.models = models_str

    config.general.datasets_file = tmpdir / "datasets.yaml"
    config.general.models_file = tmpdir / "models.yaml"
    analysis.write_datasets()
    analysis = Analysis(config)
    analysis.read_datasets()
    assert len(analysis.datasets.models) == 2
    assert analysis.models.names == ["source", "stacked-bkg"]

    analysis.models[0].parameters["index"].value = 3
    analysis.write_models()
    analysis = Analysis(config)
    analysis.read_datasets()
    assert len(analysis.datasets.models) == 2
    assert analysis.models.names == ["source", "stacked-bkg"]
    assert analysis.models[0].parameters["index"].value == 3



"""
Dark matter spatial and spectral models
=======================================

Convenience methods for dark matter high level analyses.

Introduction
------------

Gammapy has some convenience methods for dark matter analyses in
`gammapy.astro.darkmatter`. These include J-Factor computation and
calculation the expected gamma flux for a number of annihilation
channels. They are presented in this notebook.

The basic concepts of indirect dark matter searches, however, are not
explained. So this is aimed at people who already know what the want to
do. A good introduction to indirect dark matter searches is given for
example `here <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C/abstract>`__ (Chapter 1 and 5).

"""

######################################################################
# Setup
# -----
#
# As always, we start with some setup for the notebook, and with imports.
#

import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord
from regions import CircleSkyRegion, RectangleSkyRegion

# %matplotlib inline
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from gammapy.astro.darkmatter import (
    DarkMatterAnnihilationSpectralModel,
    DarkMatterDecaySpectralModel,
    JFactory,
    PrimaryFlux,
    profiles,
)
from gammapy.maps import WcsGeom, WcsNDMap


######################################################################
# Profiles
# --------
#
# The following dark matter profiles are currently implemented. Each model
# can be scaled to a given density at a certain distance. These parameters
# are controlled by `~gammapy.astro.darkmatter.profiles.DMProfile.LOCAL_DENSITY` and
# `~gammapy.astro.darkmatter.profiles.DMProfile.DISTANCE_GC`
#

profiles.DMProfile.__subclasses__()

for profile in profiles.DMProfile.__subclasses__():
    p = profile()
    p.scale_to_local_density()
    radii = np.logspace(-3, 2, 100) * u.kpc
    plt.plot(radii, p(radii), label=p.__class__.__name__)

plt.loglog()
plt.axvline(8.5, linestyle="dashed", color="black", label="local density")
plt.legend()
plt.show()

print("LOCAL_DENSITY:", profiles.DMProfile.LOCAL_DENSITY)
print("DISTANCE_GC:", profiles.DMProfile.DISTANCE_GC)


######################################################################
# J Factors
# ---------
#
# There are utilities to compute J-Factor maps that can serve as a basis
# to compute J-Factors for certain regions. In the following we compute a
# J-Factor annihilation map for the Galactic Centre region
#

profile = profiles.NFWProfile(r_s=20 * u.kpc)

######################################################################
# Adopt standard values used in H.E.S.S.

profiles.DMProfile.DISTANCE_GC = 8.5 * u.kpc
profiles.DMProfile.LOCAL_DENSITY = 0.39 * u.Unit("GeV / cm3")

profile.scale_to_local_density()

position = SkyCoord(0.0, 0.0, frame="galactic", unit="deg")
geom = WcsGeom.create(binsz=0.05, skydir=position, width=3.0, frame="galactic")

jfactory = JFactory(geom=geom, profile=profile, distance=profiles.DMProfile.DISTANCE_GC)
jfact = jfactory.compute_jfactor()

jfact_map = WcsNDMap(geom=geom, data=jfact.value, unit=jfact.unit)
plt.figure()
ax = jfact_map.plot(cmap="viridis", norm=LogNorm(), add_cbar=True)
plt.title(f"J-Factor [{jfact_map.unit}]")

######################################################################
# 1 deg circle usually used in H.E.S.S. analyses without the +/- 0.3 deg band around the plane

sky_reg = CircleSkyRegion(center=position, radius=1 * u.deg)
pix_reg = sky_reg.to_pixel(wcs=geom.wcs)
pix_reg.plot(ax=ax, facecolor="none", edgecolor="red", label="1 deg circle")

sky_reg_rec = RectangleSkyRegion(center=position, height=0.6 * u.deg, width=2 * u.deg)
pix_reg_rec = sky_reg_rec.to_pixel(wcs=geom.wcs)
pix_reg_rec.plot(ax=ax, facecolor="none", edgecolor="orange", label="+/- 0.3 deg band")

plt.legend()
plt.show()

######################################################################
# Note the value quoted by H.E.S.S. in `this paper <https://arxiv.org/abs/1607.08142>`__
# is :math:`2.67\times10^{21}`

total_jfact = (
    pix_reg.to_mask().multiply(jfact).sum()
    - pix_reg_rec.to_mask().multiply(jfact).sum()
)
total_jfact = (
    pix_reg.to_mask().multiply(jfact).sum()
    - pix_reg_rec.to_mask().multiply(jfact).sum()
)
print(
    "J-factor in 1 deg circle without the +/- 0.3 deg band around GC assuming a "
    f"{profile.__class__.__name__} is {total_jfact:.3g}"
)

######################################################################
# The J-Factor can also be computed for dark matter decay

jfactory = JFactory(
    geom=geom,
    profile=profile,
    distance=profiles.DMProfile.DISTANCE_GC,
    annihilation=False,
)
jfact_decay = jfactory.compute_jfactor()

jfact_map = WcsNDMap(geom=geom, data=jfact_decay.value, unit=jfact_decay.unit)
plt.figure()
ax = jfact_map.plot(cmap="viridis", norm=LogNorm(), add_cbar=True)
plt.title(f"J-Factor [{jfact_map.unit}]")

######################################################################
# 1 deg circle usually used in H.E.S.S. analyses without the +/- 0.3 deg band around the plane

sky_reg = CircleSkyRegion(center=position, radius=1 * u.deg)
pix_reg = sky_reg.to_pixel(wcs=geom.wcs)
pix_reg.plot(ax=ax, facecolor="none", edgecolor="red", label="1 deg circle")

sky_reg_rec = RectangleSkyRegion(center=position, height=0.6 * u.deg, width=2 * u.deg)
pix_reg_rec = sky_reg_rec.to_pixel(wcs=geom.wcs)
pix_reg_rec.plot(ax=ax, facecolor="none", edgecolor="orange", label="+/- 0.3 deg band")

plt.legend()
plt.show()

total_jfact_decay = (
    pix_reg.to_mask().multiply(jfact_decay).sum()
    - pix_reg_rec.to_mask().multiply(jfact_decay).sum()
)
total_jfact_decay = (
    pix_reg.to_mask().multiply(jfact_decay).sum()
    - pix_reg_rec.to_mask().multiply(jfact_decay).sum()
)
print(
    "J-factor in 1 deg circle without the +/- 0.3 deg band around GC assuming a "
    f"{profile.__class__.__name__} is {total_jfact_decay:.3g}"
)

######################################################################
# Gamma-ray spectra at production
# -------------------------------
#
# The gamma-ray spectrum per annihilation is a further ingredient for a
# dark matter analysis. The following annihilation channels are supported.
# For more info see https://arxiv.org/pdf/1012.4515.pdf
#

fluxes = PrimaryFlux(mDM="1 TeV", channel="eL")
print(fluxes.allowed_channels)

fig, axes = plt.subplots(4, 1, figsize=(4, 16))
mDMs = [0.01, 0.1, 1, 10] * u.TeV

for mDM, ax in zip(mDMs, axes):
    fluxes.mDM = mDM
    ax.set_title(rf"m$_{{\mathrm{{DM}}}}$ = {mDM}")
    ax.set_yscale("log")
    ax.set_ylabel("dN/dE")

    for channel in ["tau", "mu", "b", "Z"]:
        fluxes = PrimaryFlux(mDM=mDM, channel=channel)
        fluxes.channel = channel
        fluxes.plot(
            energy_bounds=[mDM / 100, mDM],
            ax=ax,
            label=channel,
            yunits=u.Unit("1/GeV"),
        )

axes[0].legend()
plt.subplots_adjust(hspace=0.9)
plt.show()


######################################################################
# Flux maps for annihilation
# --------------------------
#
# Finally flux maps can be produced like this:
#

channel = "Z"
massDM = 10 * u.TeV
diff_flux = DarkMatterAnnihilationSpectralModel(mass=massDM, channel=channel)
int_flux = (
    jfact * diff_flux.integral(energy_min=0.1 * u.TeV, energy_max=10 * u.TeV)
).to("cm-2 s-1")

flux_map = WcsNDMap(geom=geom, data=int_flux.value, unit="cm-2 s-1")
plt.figure()
ax = flux_map.plot(cmap="viridis", norm=LogNorm(), add_cbar=True)
plt.title(
    f"Flux [{int_flux.unit}]\n m$_{{DM}}$={fluxes.mDM.to('TeV')}, channel={fluxes.channel}"
)

plt.show()


######################################################################
# Flux maps for decay
# -------------------
#
# Finally flux maps for decay can be produced like this:
#

channel = "Z"
massDM = 10 * u.TeV
diff_flux = DarkMatterDecaySpectralModel(mass=massDM, channel=channel)
int_flux = (
    jfact_decay * diff_flux.integral(energy_min=0.1 * u.TeV, energy_max=10 * u.TeV)
).to("cm-2 s-1")

flux_map = WcsNDMap(geom=geom, data=int_flux.value, unit="cm-2 s-1")
plt.figure()
ax = flux_map.plot(cmap="viridis", norm=LogNorm(), add_cbar=True)
plt.title(
    f"Flux [{int_flux.unit}]\n m$_{{DM}}$={fluxes.mDM.to('TeV')}, channel={fluxes.channel}"
)

plt.show()
